#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.constants import *
from reclaimer.util import fourcc_to_int


# These are the 4 fourCC that I've found in tag headers
# and this is what I believe the order of their age is.
engine_id_to_name = dict(
    BLM_="halo_2",
    LAMB="halo_2_old",
    MLAB="halo_2_older",
    ambl="halo_2_oldest",
    )

# Here are some of the differences between engine versions:
# ambl: the tbfd struct was 12 bytes instead of 16, with the second and
#       third UInt32 replaced by UInt16, which are the version and bcount
# MLAB: the ascii_str32 instances were replaced with ascii_str_varlen
# LAMB: ???
# BLM!: newest version.

# DO NOT CHANGE THE ORDER OF THESE
HALO2_MAP_TYPES = ("local", "mainmenu", "shared", "single_player_shared")


# bitmap formats
PALLETIZED_FORMATS = (FORMAT_P8_BUMP, FORMAT_P8)


# maps tag class four character codes(fccs) in
# their string encoding to their int encoding.
h2_tag_class_fcc_to_be_int = {}
h2_tag_class_fcc_to_le_int = {}
# maps tag class four character codes(fccs) in
# their int encoding to their string encoding.
h2_tag_class_be_int_to_fcc = {}
h2_tag_class_le_int_to_fcc = {}

# maps tag class four character codes to the tags file extension
# 120 classes, 97 of which are NOT marked with OLD?
h2_tag_class_fcc_to_ext = {
    "adlg": "ai_dialogue_globals",
    "mdlg": "ai_mission_dialogue",
    "ant!": "antenna",
    "bipd": "biped",
    "bitm": "bitmap",
    "bsdt": "breakable_surface",
    "$#!+": "cache_file_sound",  # OLD?
    "trak": "camera_track",
    "devo": "cellular_automata",
    "whip": "cellular_automata2d",
    "char": "character",
    "gldf": "chocolate_mountain",
    "clwd": "cloth",
    "coll": "collision_model",
    "coln": "colony",
    "colo": "color_table",
    "cont": "contrail",
    "bloc": "crate",
    "crea": "creature",
    "jpt!": "damage_effect",
    "deca": "decal",
    "DECR": "decorator_set",
    "DECP": "decorators",
    "dobc": "detail_object_collection",
    "devi": "device",
    "ctrl": "device_control",
    "lifi": "device_light_fixture",
    "mach": "device_machine",
    "udlg": "dialogue",
    "effe": "effect",
    "eqip": "equipment",
    "garb": "garbage",
    "matg": "globals",
    "grhi": "grenade_hud_interface",  # OLD?
    "hudg": "hud_globals",
    "hmt ": "hud_message_text",
    "hud#": "hud_number",
    "item": "item",
    "itmc": "item_collection",
    "lens": "lens_flare",
    "ligh": "light",
    "MGS2": "light_volume",
    "tdtl": "liquid",
    "foot": "material_effects",
    "mpdt": "material_physics",
    "metr": "meter",
    "hlmt": "model",
    "jmad": "model_animation_graph",
    "mcsr": "mouse_cursor_definition",
    "unic": "multilingual_unicode_string_list",
    "mulg": "multiplayer_globals",
    "mply": "multiplayer_scenario_description",
    "goof": "multiplayer_variant_settings_interface_definition",
    "nhdt": "new_hud_definition",
    "obje": "object",
    'pctl': "particle_system",  # still in use in the halo 2 alpha
    "part": "particle_old",     # still in use in the halo 2 alpha
    "prt3": "particle",
    "PRTM": "particle_model",
    "pmov": "particle_physics",
    "fpch": "patchy_fog",
    "phys": "physics",
    "phmo": "physics_model",
    "pixl": "pixel_shader",
    "fog ": "planar_fog",
    "pphy": "point_physics",
    "proj": "projectile",
    "mode": "render_model",
    "sbsp": "scenario_structure_bsp",
    "ltmp": "scenario_structure_lightmap",
    "scnr": "scenario",
    "ai**": "scenario_ai_resource",  # OLD?
    "*ipd": "scenario_bipeds_resource",  # OLD?
    "cin*": "scenario_cinematics_resource",  # OLD?
    "clu*": "scenario_cluster_data_resource",  # OLD?
    "/**/": "scenario_comments_resource",  # OLD?
    "*rea": "scenario_creature_resource",  # OLD?
    "dec*": "scenario_decals_resource",  # OLD?
    "dc*s": "scenario_decorators_resource",  # OLD?
    "dgr*": "scenario_devices_resource",  # OLD?
    "*qip": "scenario_equipment_resource",  # OLD?
    "hsc*": "scenario_hs_source_file",  # OLD?
    "*cen": "scenario_scenery_resource",  # OLD?
    "*sce": "scenario_sound_scenery_resource",  # OLD?
    "sslt": "scenario_structure_lighting_resource",  # OLD?
    "*igh": "scenario_lights_resource",  # OLD?
    "trg*": "scenario_trigger_volumes_resource",  # OLD?
    "*ehi": "scenario_vehicles_resource",  # OLD?
    "*eap": "scenario_weapons_resource",  # OLD?
    "scen": "scenery",
    "egor": "screen_effect",
    "shad": "shader",
    "stem": "shader_template",
    "slit": "shader_light_response",
    "spas": "shader_pass",
    "sky ": "sky",
    "snd!": "sound",
    "ugh!": "sound_cache_file_gestalt",
    "sncl": "sound_classes",
    "spk!": "sound_dialogue_constants",
    "<fx>": "sound_effect_template",
    "sfx+": "sound_effect_collection",
    "snde": "sound_environment",
    "lsnd": "sound_looping",
    "snmx": "sound_mix",
    "ssce": "sound_scenery",
    "BooM": "stereo_system",
    "styl": "style",
    "sily": "text_value_pair_definition",
    "unit": "unit",
    "unhi": "unit_hud_interface",  # OLD?
    "wgtz": "user_interface_globals_definition",
    "skin": "user_interface_list_skin_definition",
    "wgit": "user_interface_screen_widget_definition",
    "wigl": "user_interface_shared_globals_definition",
    "vehi": "vehicle",
    "vehc": "vehicle_collection",
    "vrtx": "vertex_shader",
    "weap": "weapon",
    "wphi": "weapon_hud_interface",  # OLD?
    "weat": "weather_system",
    "wind": "wind",
    }

for tag_cls in h2_tag_class_fcc_to_ext:
    h2_tag_class_fcc_to_be_int[tag_cls] = fourcc_to_int(tag_cls)
    h2_tag_class_be_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls
    h2_tag_class_fcc_to_le_int[tag_cls] = fourcc_to_int(tag_cls)
    h2_tag_class_le_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Deadcanadian
# 	naming whatever i can figure out
# revision: 3		author: Lord Zedd
# 	Yeah Buddy.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

wezr_capture_the_flag_variant_game_mode = (
    "multiple",
    "single",
    "neutral",
    )

wezr_capture_the_flag_variant_home_flag_waypoint = (
    "unknown_1",
    "unknown_2",
    "unknown_3",
    "not_in_single",
    )

wezr_capture_the_flag_variant_respawn_on_capture = (
    "disabled",
    "on_ally_capture",
    "on_enemy_capture",
    "on_any_capture",
    )

wezr_infection_variant_next_zombie = (
    "most_points",
    "first_infected",
    "unchanged",
    "random",
    )

wezr_infection_variant_safe_havens = (
    "none",
    "random",
    "sequence",
    )

wezr_juggernaut_variant_first_juggernaut = (
    "random",
    "first_kill",
    "first_death",
    )

wezr_juggernaut_variant_goal_zone_movement = (
    "no_movement",
    "after_10_seconds",
    "after_15_seconds",
    "after_30_seconds",
    "after_1_minute",
    "after_2_minutes",
    "after_3_minutes",
    "after_4_minutes",
    "after_5_minutes",
    "on_arrival",
    "on_new_juggernaut",
    )

wezr_juggernaut_variant_next_juggernaut = (
    "killer",
    "killed",
    "unchanged",
    "random",
    )

wezr_king_of_the_hill_variant_hill_movement = (
    "no_movement",
    "after_10_seconds",
    "after_15_seconds",
    "after_30_seconds",
    "after_1_minute",
    "after_2_minutes",
    "after_3_minutes",
    "after_4_minutes",
    "after_5_minutes",
    )

wezr_king_of_the_hill_variant_hill_movement_order = (
    "random",
    "sequence",
    )

wezr_king_of_the_hill_variant_team_scoring = (
    "sum",
    "minimum",
    "maximum",
    "default",
    )

wezr_sandbox_editor_variant_edit_mode = (
    "everyone",
    "leader_only",
    )

wezr_slayer_variant_general_setting_round_resets = (
    "nothing",
    "players_only",
    "everything",
    )

wezr_slayer_variant_team_scoring = (
    "sum_of_team",
    "minimum_score",
    "maximum_score",
    "default",
    )

wezr_trait_profile_appearance_active_camo = (
    "unchanged",
    "disabled",
    "bad_camo",
    "poor_camo",
    "good_camo",
    )

wezr_trait_profile_appearance_aura = (
    "unchanged",
    "disabled",
    "team",
    "black",
    "white",
    )

wezr_trait_profile_appearance_forced_color = (
    "unchanged",
    "off",
    "red",
    "blue",
    "green",
    "orange",
    "purple",
    "gold",
    "brown",
    "pink",
    "white",
    "black",
    "zombie",
    "pink_unused",
    )

wezr_trait_profile_appearance_waypoint = (
    "unchanged",
    "none",
    "visible_to_allies",
    "visible_to_everyone",
    )

wezr_trait_profile_movement_player_gravity = (
    "unchanged",
    "_50",
    "_75",
    "_100",
    "_150",
    "_200",
    )

wezr_trait_profile_movement_player_speed = (
    "unchanged",
    "_25",
    "_50",
    "_75",
    "_90",
    "_100",
    "_110",
    "_125",
    "_150",
    "_200",
    "_300",
    )

wezr_trait_profile_movement_vehicle_use = (
    "unchanged",
    "none",
    "passenger_only",
    "full_use",
    )

wezr_trait_profile_sensor_motion_tracker_mode = (
    "unchanged",
    "disabled",
    "ally_movement",
    "player_movement",
    "player_locations",
    )

wezr_trait_profile_sensor_motion_tracker_range = (
    "unchanged",
    "_10m",
    "_15m",
    "_25m",
    "_50m",
    "_75m",
    "_100m",
    "_150m",
    )

wezr_trait_profile_shields_and_health_damage_resistance = (
    "unchanged",
    "_10",
    "_50",
    "_90",
    "_100",
    "_110",
    "_150",
    "_200",
    "_300",
    "_500",
    "_1000",
    "_2000",
    "invulnerable",
    )

wezr_trait_profile_shields_and_health_headshot_immunity = (
    "unchanged",
    "enabled",
    "disabled",
    )

wezr_trait_profile_shields_and_health_shield_multiplier = (
    "unchanged",
    "no_shields",
    "normal_shields",
    "_2x_overshields",
    "_3x_overshields",
    "_4x_overshields",
    )

wezr_trait_profile_shields_and_health_shield_recharge_rate = (
    "unchanged",
    "neg_25",
    "neg_10",
    "neg_5",
    "_0",
    "_50",
    "_90",
    "_100",
    "_110",
    "_200",
    )

wezr_trait_profile_shields_and_health_shield_vampirism = (
    "unchanged",
    "disabled",
    "_10",
    "_25",
    "_50",
    "_100",
    )

wezr_trait_profile_weapons_and_damage_damage_modifier = (
    "unchanged",
    "_0",
    "_25",
    "_50",
    "_75",
    "_90",
    "_100",
    "_110",
    "_125",
    "_150",
    "_200",
    "_300",
    "instant_kill",
    )

wezr_trait_profile_weapons_and_damage_grenade_count = (
    "unchanged",
    "map_default",
    "none",
    )

wezr_trait_profile_weapons_and_damage_infinite_ammo = (
    "unchanged",
    "disabled",
    "enabled",
    )

wezr_vip_variant_goal_zone_movement = (
    "no_movement",
    "after_10_seconds",
    "after_15_seconds",
    "after_30_seconds",
    "after_1_minute",
    "after_2_minutes",
    "after_3_minutes",
    "after_4_minutes",
    "after_5_minutes",
    "on_arrival",
    "on_new_vip",
    )

wezr_vip_variant_next_vip = (
    "random",
    "unknown",
    "next_death",
    "unchanged",
    )


wezr_trait_profile_shields_and_health = Struct("shields_and_health",
    SEnum8("damage_resistance", *wezr_trait_profile_shields_and_health_damage_resistance),
    SEnum8("shield_multiplier", *wezr_trait_profile_shields_and_health_shield_multiplier),
    SEnum8("shield_recharge_rate", *wezr_trait_profile_shields_and_health_shield_recharge_rate),
    SEnum8("headshot_immunity", *wezr_trait_profile_shields_and_health_headshot_immunity),
    SEnum8("shield_vampirism", *wezr_trait_profile_shields_and_health_shield_vampirism),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


wezr_trait_profile_weapons_and_damage = Struct("weapons_and_damage",
    SEnum8("damage_modifier", *wezr_trait_profile_weapons_and_damage_damage_modifier),
    SEnum8("grenade_regeneration", *wezr_trait_profile_shields_and_health_headshot_immunity),
    SEnum8("weapon_pickup", *wezr_trait_profile_shields_and_health_headshot_immunity),
    SEnum8("infinite_ammo", *wezr_trait_profile_weapons_and_damage_infinite_ammo),
    h3_string_id("primary_weapon"),
    h3_string_id("secondary_weapon"),
    SEnum16("grenade_count", *wezr_trait_profile_weapons_and_damage_grenade_count),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_trait_profile_movement = Struct("movement",
    SEnum8("player_speed", *wezr_trait_profile_movement_player_speed),
    SEnum8("player_gravity", *wezr_trait_profile_movement_player_gravity),
    SEnum8("vehicle_use", *wezr_trait_profile_movement_vehicle_use),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=4
    )


wezr_trait_profile_appearance = Struct("appearance",
    SEnum8("active_camo", *wezr_trait_profile_appearance_active_camo),
    SEnum8("waypoint", *wezr_trait_profile_appearance_waypoint),
    SEnum8("aura", *wezr_trait_profile_appearance_aura),
    SEnum8("forced_color", *wezr_trait_profile_appearance_forced_color),
    ENDIAN=">", SIZE=4
    )


wezr_trait_profile_sensor = Struct("sensor",
    SEnum32("motion_tracker_mode", *wezr_trait_profile_sensor_motion_tracker_mode),
    SEnum32("motion_tracker_range", *wezr_trait_profile_sensor_motion_tracker_range),
    ENDIAN=">", SIZE=8
    )


wezr_trait_profile = Struct("trait_profile",
    h3_string_id("name"),
    h3_reflexive("shields_and_health", wezr_trait_profile_shields_and_health),
    h3_reflexive("weapons_and_damage", wezr_trait_profile_weapons_and_damage),
    h3_reflexive("movement", wezr_trait_profile_movement),
    h3_reflexive("appearance", wezr_trait_profile_appearance),
    h3_reflexive("sensors", wezr_trait_profile_sensor),
    ENDIAN=">", SIZE=64
    )


wezr_slayer_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_slayer_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_slayer_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_slayer_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_slayer_variant = Struct("slayer_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_slayer_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_slayer_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_slayer_variant_social_setting),
    h3_reflexive("map_overrides", wezr_slayer_variant_map_override),
    SEnum16("team_scoring", *wezr_slayer_variant_team_scoring),
    SInt16("points_to_win"),
    SInt8("kill_points"),
    SInt8("assist_points"),
    SInt8("death_points"),
    SInt8("suicide_points"),
    SInt8("betrayal_points"),
    SInt8("leader_kill_bonus"),
    SInt8("elimination_bonus"),
    SInt8("assassination_bonus"),
    SInt8("headshot_bonus"),
    SInt8("beatdown_bonus"),
    SInt8("sticky_bonus"),
    SInt8("splatter_bonus"),
    SInt8("spree_bonus"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_string_id("leader_trait_profile"),
    ENDIAN=">", SIZE=80
    )


wezr_oddball_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_oddball_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_oddball_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_oddball_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_oddball_variant = Struct("oddball_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_oddball_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_oddball_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_oddball_variant_social_setting),
    h3_reflexive("map_overrides", wezr_oddball_variant_map_override),
    Bool32("flags",
        "autopickup_enabled",
        "ball_effect_enabled",
        ),
    SEnum16("team_scoring", *wezr_slayer_variant_team_scoring),
    SInt16("points_to_win"),
    SInt8("carrying_points"),
    SInt8("kill_points"),
    SInt8("ball_kill_points"),
    SInt8("ball_carrier_kill_points"),
    SInt8("ball_count"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    SInt16("initial_ball_delay"),
    SInt16("ball_respawn_delay"),
    h3_string_id("ball_carrier_trait_profile"),
    ENDIAN=">", SIZE=80
    )


wezr_capture_the_flag_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_capture_the_flag_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_capture_the_flag_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_capture_the_flag_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_capture_the_flag_variant = Struct("capture_the_flag_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_capture_the_flag_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_capture_the_flag_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_capture_the_flag_variant_social_setting),
    h3_reflexive("map_overrides", wezr_capture_the_flag_variant_map_override),
    Bool32("flags",
        "flag_at_home_to_score",
        ),
    SEnum16("home_flag_waypoint", *wezr_capture_the_flag_variant_home_flag_waypoint),
    SEnum16("game_mode", *wezr_capture_the_flag_variant_game_mode),
    SEnum16("respawn_on_capture", *wezr_capture_the_flag_variant_respawn_on_capture),
    SInt16("flag_return_time"),
    SInt16("sudden_death_time"),
    SInt16("score_to_win"),
    SInt16("flag_reset_time"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    h3_string_id("flag_carrier_trait_profile"),
    ENDIAN=">", SIZE=80
    )


wezr_assault_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_assault_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_assault_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_assault_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_assault_variant = Struct("assault_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_assault_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_assault_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_assault_variant_social_setting),
    h3_reflexive("map_overrides", wezr_assault_variant_map_override),
    Bool32("flags",
        "reset_on_disarm",
        ),
    SEnum16("respawn_on_capture", *wezr_capture_the_flag_variant_respawn_on_capture),
    SEnum16("game_mode", *wezr_capture_the_flag_variant_game_mode),
    SEnum16("enemy_bomb_waypoint", *wezr_capture_the_flag_variant_home_flag_waypoint),
    SInt16("sudden_death_time"),
    SInt16("detonations_to_win"),
    SInt16("bomb_reset_time"),
    SInt16("bomb_arming_time"),
    SInt16("bomb_disarming_time"),
    SInt16("bomb_fuse_time"),
    SInt16("unknown", VISIBLE=False),
    h3_string_id("bomb_carrier_trait_profile"),
    h3_string_id("unknown_trait_profile"),
    ENDIAN=">", SIZE=88
    )


wezr_infection_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_infection_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_infection_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_infection_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_infection_variant = Struct("infection_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_infection_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_infection_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_infection_variant_social_setting),
    h3_reflexive("map_overrides", wezr_infection_variant_map_override),
    Bool32("flags",
        "respawn_on_haven_move",
        ),
    SEnum16("safe_havens", *wezr_infection_variant_safe_havens),
    SEnum16("next_zombie", *wezr_infection_variant_next_zombie),
    SInt16("initial_zombie_count"),
    SInt16("safe_haven_movement_time"),
    SInt8("zombie_kill_points"),
    SInt8("infection_points"),
    SInt8("safe_haven_arrival_points"),
    SInt8("suicide_points"),
    SInt8("betrayal_points"),
    SInt8("last_man_standing_bonus"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    h3_string_id("zombie_trait_profile"),
    h3_string_id("alpha_zombie_trait_profile"),
    h3_string_id("on_haven_trait_profile"),
    h3_string_id("last_human_trait_profile"),
    ENDIAN=">", SIZE=92
    )


wezr_king_of_the_hill_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_king_of_the_hill_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_king_of_the_hill_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_king_of_the_hill_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_king_of_the_hill_variant = Struct("king_of_the_hill_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_king_of_the_hill_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_king_of_the_hill_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_king_of_the_hill_variant_social_setting),
    h3_reflexive("map_overrides", wezr_king_of_the_hill_variant_map_override),
    Bool32("flags",
        "opaque_hill",
        ),
    SInt16("score_to_win"),
    SEnum16("team_scoring", *wezr_king_of_the_hill_variant_team_scoring),
    SEnum16("hill_movement", *wezr_king_of_the_hill_variant_hill_movement),
    SEnum16("hill_movement_order", *wezr_king_of_the_hill_variant_hill_movement_order),
    SInt8("on_hill_points"),
    SInt8("uncontested_control_points"),
    SInt8("off_hill_points"),
    SInt8("kill_points"),
    h3_string_id("on_hill_trait_profile"),
    ENDIAN=">", SIZE=76
    )


wezr_territories_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_territories_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_territories_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_territories_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_territories_variant = Struct("territories_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_territories_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_territories_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_territories_variant_social_setting),
    h3_reflexive("map_overrides", wezr_territories_variant_map_override),
    Bool32("flags",
        "one_sided",
        "lock_after_first_capture",
        ),
    SEnum16("respawn_on_capture", *wezr_capture_the_flag_variant_respawn_on_capture),
    SInt16("territory_capture_time"),
    SInt16("sudden_death_time"),
    SInt16("unknown", VISIBLE=False),
    h3_string_id("defender_trait_profile"),
    h3_string_id("attacker_trait_profile"),
    ENDIAN=">", SIZE=76
    )


wezr_juggernaut_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_juggernaut_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_juggernaut_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_juggernaut_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_juggernaut_variant = Struct("juggernaut_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_juggernaut_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_juggernaut_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_juggernaut_variant_social_setting),
    h3_reflexive("map_overrides", wezr_juggernaut_variant_map_override),
    Bool32("flags",
        "allied_against_juggernaut",
        "respawn_on_lone_juggernaut",
        "goal_zones_enabled",
        ),
    SEnum16("first_juggernaut", *wezr_juggernaut_variant_first_juggernaut),
    SEnum16("next_juggernaut", *wezr_juggernaut_variant_next_juggernaut),
    SEnum16("goal_zone_movement", *wezr_juggernaut_variant_goal_zone_movement),
    SEnum16("goal_zone_order", *wezr_king_of_the_hill_variant_hill_movement_order),
    SInt16("score_to_win"),
    SInt8("kill_points"),
    SInt8("takedown_points"),
    SInt8("kill_as_juggernaut_points"),
    SInt8("goal_arrival_points"),
    SInt8("suicide_points"),
    SInt8("betrayal_points"),
    SInt8("next_juggernaut_delay"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_string_id("juggernaut_trait_profile"),
    ENDIAN=">", SIZE=84
    )


wezr_vip_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_vip_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_vip_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_vip_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_vip_variant = Struct("vip_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_vip_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_vip_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_vip_variant_social_setting),
    h3_reflexive("map_overrides", wezr_vip_variant_map_override),
    Bool32("flags",
        "single_vip",
        "goal_zones_enabled",
        "end_round_on_vip_death",
        ),
    SInt16("score_to_win"),
    SEnum16("next_vip", *wezr_vip_variant_next_vip),
    SEnum16("goal_zone_movement", *wezr_vip_variant_goal_zone_movement),
    SEnum16("goal_zone_movement_order", *wezr_king_of_the_hill_variant_hill_movement_order),
    SInt8("kill_points"),
    SInt8("vip_takedown_points"),
    SInt8("kill_as_vip_points"),
    SInt8("vip_death_points"),
    SInt8("goal_arrival_points"),
    SInt8("suicide_points"),
    SInt8("vip_betrayal_points"),
    SInt8("betrayal_points"),
    SInt8("vip_proximity_trait_radius"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_string_id("vip_team_trait_profile"),
    h3_string_id("vip_proximity_trait_profile"),
    h3_string_id("vip_trait_profile"),
    ENDIAN=">", SIZE=92
    )


wezr_sandbox_editor_variant_general_setting = Struct("general_setting",
    Bool32("flags",
        "teams_enabled",
        ),
    SInt8("time_limit"),
    SInt8("number_of_rounds"),
    SInt8("early_victory_win_count"),
    SEnum8("round_resets", *wezr_slayer_variant_general_setting_round_resets),
    ENDIAN=">", SIZE=8
    )


wezr_sandbox_editor_variant_respawn_setting = Struct("respawn_setting",
    Bool16("flags",
        "inherit_respawn_time",
        "respawn_with_team",
        "respawn_at_location",
        "respawn_on_kills",
        ),
    SInt8("lives_per_round"),
    SInt8("shared_team_lives"),
    UInt8("respawn_time"),
    UInt8("suicide_penalty"),
    UInt8("betrayal_penalty"),
    UInt8("respawn_time_growth"),
    h3_string_id("respawn_trait_profile"),
    SInt8("respawn_trait_duration"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


wezr_sandbox_editor_variant_social_setting = Struct("social_setting",
    Bool32("flags",
        "observers_enabled",
        "team_changing_enabled",
        "balanced_team_changing",
        "friendly_fire_enabled",
        "betrayal_booting_enabled",
        "enemy_voice_enabled",
        "open_channel_voice_enabled",
        "dead_player_voice_enabled",
        ),
    ENDIAN=">", SIZE=4
    )


wezr_sandbox_editor_variant_map_override = Struct("map_override",
    Bool32("flags",
        "grenades_on_map",
        "indestructable_vehicles",
        ),
    h3_string_id("base_player_trait_profile"),
    h3_string_id("weapon_set"),
    h3_string_id("vehicle_set"),
    h3_string_id("overshield_trait_profile"),
    h3_string_id("active_camo_trait_profile"),
    h3_string_id("custom_powerup_trait_profile"),
    SInt8("overshield_trait_duration"),
    SInt8("active_camo_trait_duration"),
    SInt8("custom_powerup_trait_duration"),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wezr_sandbox_editor_variant = Struct("sandbox_editor_variant",
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("general_settings", wezr_sandbox_editor_variant_general_setting),
    h3_reflexive("respawn_settings", wezr_sandbox_editor_variant_respawn_setting),
    h3_reflexive("social_settings", wezr_sandbox_editor_variant_social_setting),
    h3_reflexive("map_overrides", wezr_sandbox_editor_variant_map_override),
    Bool32("flags",
        "open_channel_voice_enabled",
        ),
    SEnum16("edit_mode", *wezr_sandbox_editor_variant_edit_mode),
    SInt16("editor_respawn_time"),
    h3_string_id("editor_trait_profile"),
    ENDIAN=">", SIZE=68
    )


wezr_body = Struct("tagdata",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    h3_reflexive("trait_profiles", wezr_trait_profile),
    h3_reflexive("slayer_variants", wezr_slayer_variant),
    h3_reflexive("oddball_variants", wezr_oddball_variant),
    h3_reflexive("capture_the_flag_variants", wezr_capture_the_flag_variant),
    h3_reflexive("assault_variants", wezr_assault_variant),
    h3_reflexive("infection_variants", wezr_infection_variant),
    h3_reflexive("king_of_the_hill_variants", wezr_king_of_the_hill_variant),
    h3_reflexive("territories_variants", wezr_territories_variant),
    h3_reflexive("juggernaut_variants", wezr_juggernaut_variant),
    h3_reflexive("vip_variants", wezr_vip_variant),
    h3_reflexive("sandbox_editor_variants", wezr_sandbox_editor_variant),
    ENDIAN=">", SIZE=136
    )


def get():
    return wezr_def

wezr_def = TagDef("wezr",
    h3_blam_header('wezr'),
    wezr_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wezr"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import cos

from reclaimer.hek.defs.objs.tag import HekTag

class ActrTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)
        perception = self.data.tagdata.perception
        looking = self.data.tagdata.looking

        perception.inv_combat_perception_time = 0
        perception.inv_guard_perception_time = 0
        perception.inv_non_combat_perception_time = 0

        if perception.combat_perception_time:
            perception.inv_combat_perception_time = 1 / perception.combat_perception_time

        if perception.guard_perception_time:
            perception.inv_guard_perception_time = 1 / perception.guard_perception_time

        if perception.non_combat_perception_time:
            perception.inv_non_combat_perception_time = 1 / perception.non_combat_perception_time

        perception.inv_combat_perception_time /= 30
        perception.inv_guard_perception_time /= 30
        perception.inv_non_combat_perception_time /= 30

        for i in range(2):
            looking.cosine_maximum_aiming_deviation[i]  = cos(looking.maximum_aiming_deviation[i])
            looking.cosine_maximum_looking_deviation[i] = cos(looking.maximum_looking_deviation[i])



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import unpack

from supyr_struct.defs import sanitizers
from supyr_struct.field_types import *
from supyr_struct.blocks import EnumBlock
from reclaimer.field_type_methods import encode_tag_ref_str,\
     tag_ref_str_sizecalc, tag_ref_str_parser, tag_ref_str_serializer,\
     decode_raw_string, decode_string, decoder_wrapper, encoder_wrapper,\
     tag_cstring_parser, rawdata_ref_parser, reflexive_parser, encode_string,\
     utf_sizecalc, delim_utf_sizecalc, sizecalc_wrapper, len_sizecalc


'''These are varients of the standard FieldTypes that have been
slightly modified based on how Halo needs to utilize them.'''
StrTagRef = FieldType(
    base=StrLatin1, name="StrTagRef",
    parser=tag_ref_str_parser, serializer=tag_ref_str_serializer,
    encoder=encode_tag_ref_str, sizecalc=tag_ref_str_sizecalc)
FlStrUTF16Data = FieldType(
    base=StrUtf16, name="StrUTF16Data",
    enc={">": StrUtf16.little.enc, "<": StrUtf16.little.enc},
    decoder=decode_raw_string, sizecalc=utf_sizecalc)
FlStrUTF16 = FieldType(
    base=StrUtf16, name="StrUTF16",
    enc={">": StrUtf16.little.enc, "<": StrUtf16.little.enc},
    decoder=decode_string, sizecalc=delim_utf_sizecalc)
CStrTagRef = FieldType(
    base=CStrLatin1, name="CStrTagRef", parser=tag_cstring_parser
    )

#forces little endian integers and float
FlUInt16 = FieldType(
    base=UInt16.little, name="FlUInt16", enc=UInt16.little.enc)
FlUInt32 = FieldType(
    base=UInt32.little, name="FlUInt32", enc=UInt32.little.enc)

FlUEnum16 = FieldType(
    base=UEnum16.little, name="FlUEnum16", enc=UEnum16.little.enc)
FlUEnum32 = FieldType(
    base=UEnum32.little, name="FlUEnum32", enc=UEnum32.little.enc)

FlBool16 = FieldType(
    base=Bool16.little, name="FlBool16", enc=Bool16.little.enc)
FlBool32 = FieldType(
    base=Bool32.little, name="FlBool32", enc=Bool32.little.enc)

FlSInt16 = FieldType(
    base=SInt16.little, name="FlSInt16", enc=SInt16.little.enc)
FlSInt32 = FieldType(
    base=SInt32.little, name="FlSInt32", enc=SInt32.little.enc)

FlSEnum16 = FieldType(
    base=SEnum16.little, name="FlSEnum16", enc=SEnum16.little.enc)
FlSEnum32 = FieldType(
    base=SEnum32.little, name="FlSEnum32", enc=SEnum32.little.enc)

FlFloat = FieldType(base=Float.little, name="FlFloat", enc=Float.little.enc)

'''These FieldTypes exist to easily identify where in a tag
that raw data refs, reflexives, and tag references exist.'''
RawdataRef   = FieldType(base=Struct,  name="RawdataRef", parser=rawdata_ref_parser)
Reflexive    = FieldType(base=QStruct, name="Reflexive", parser=reflexive_parser)
RawReflexive = FieldType(base=Reflexive, name="RawReflexive")
TagRef       = FieldType(base=Struct, name="TagRef")

ZoneAsset = FieldType(base=QStruct, name="ZoneAsset")
StringID  = FieldType(base=QStruct, name="StringID")
TagIndex  = FieldType(base=Array, name="TagIndex")

StrLatin1Enum = FieldType(
    base=StrRawLatin1, name="StrLatin1Enum", is_block=True, is_data=True,
    sizecalc=sizecalc_wrapper(len_sizecalc), data_cls=str,
    decoder=decoder_wrapper(decode_string),
    encoder=encoder_wrapper(encode_string),
    node_cls=EnumBlock, sanitizer=sanitizers.enum_sanitizer)

del EnumBlock
del encode_tag_ref_str
del tag_ref_str_sizecalc
del tag_ref_str_parser
del tag_ref_str_serializer
del decode_raw_string
del decode_string
del decoder_wrapper
del encoder_wrapper
del tag_cstring_parser
del rawdata_ref_parser
del reflexive_parser
del encode_string
del utf_sizecalc
del delim_utf_sizecalc
del sizecalc_wrapper
del len_sizecalc



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

button_icon = Struct("button_icon",
    SInt16("sequence_index"),
    SInt16("width_offset"),
    QStruct("offset_from_reference_corner",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    #QStruct("override_icon_color", INCLUDE=argb_byte),
    UInt32("override_icon_color", INCLUDE=argb_uint32),
    SInt8("frame_rate", MIN=0, MAX=30, UNIT_SCALE=per_sec_unit_scale),
    Bool8("flags",
        "use_text_from_string_list_instead",
        "override_default_color",
        "width_offset_is_absolute_icon_width",
        ),
    SInt16("text_index"),
    SIZE=16
    )

waypoint_arrow = Struct("waypoint_arrow",
    ascii_str32("name"),

    Pad(8),
    #QStruct("color", INCLUDE=xrgb_byte),
    UInt32("color", INCLUDE=xrgb_uint32),
    Float("opacity"),
    Float("translucency"),
    SInt16("on_screen_sequence_index"),
    SInt16("off_screen_sequence_index"),
    SInt16("occluded_sequence_index"),

    Pad(18),
    Bool32("flags",
        "dont_rotate_when_pointing_offscreen"
        ),
    SIZE=104
    )


messaging_parameters = Struct("messaging_parameters",
    SEnum16("anchor", *hud_anchors),

    Pad(34),
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    dependency("single_player_font", "font"),
    dependency("multi_player_font", "font"),
    float_sec("up_time"),
    float_sec("fade_time"),
    QStruct("icon_color", INCLUDE=argb_float),
    QStruct("text_color", INCLUDE=argb_float),
    Float("text_spacing"),
    dependency("item_message_text", "ustr"),
    dependency("icon_bitmap", "bitm"),
    dependency("alternate_icon_text", "ustr"),
    SIZE=196
    )

hud_help_text_color = Struct("hud_help_text_color",
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    SIZE=32
    )

objective_colors = Struct("objective_colors",
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    SInt16("uptime_ticks", UNIT_SCALE=sec_unit_scale),
    SInt16("fade_ticks", UNIT_SCALE=sec_unit_scale),
    SIZE=32
    )

waypoint_parameters = Struct("waypoint_parameters",
    Float("top_offset"),
    Float("bottom_offset"),
    Float("left_offset"),
    Float("right_offset"),

    Pad(32),
    dependency("arrow_bitmaps", "bitm"),
    SIZE=64
    )

hud_globals = Struct("hud_globals",
    Float("hud_scale_in_multiplayer"),

    Pad(256),
    dependency("default_weapon_hud", "wphi"),
    Float("motion_sensor_range"),
    Float("motion_sensor_velocity_sensitivity"),
    Float("motion_sensor_scale", DEFAULT=32.0),  # DONT TOUCH(why?)
    QStruct("default_chapter_title_bounds",
        SInt16("t"), SInt16("l"), SInt16("b"), SInt16("r"), ORIENT='h'
        ),
    SIZE=340
    )

hud_damage_indicators = Struct("hud_damage_indicators",
    SInt16("top_offset"),
    SInt16("bottom_offset"),
    SInt16("left_offset"),
    SInt16("right_offset"),

    Pad(32),
    dependency("indicator_bitmap", "bitm"),
    SInt16("sequence_index"),
    SInt16("multiplayer_sequence_index"),
    #QStruct("color", INCLUDE=argb_byte),
    UInt32("color", INCLUDE=argb_uint32),
    SIZE=80
    )

time_running_out = Struct("time_running_out_flash_color",
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    SIZE=32
    )

time_out = Struct("time_out_flash_color",
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    SIZE=32
    )

misc_hud_crap = Struct("misc_hud_crap",
    SInt16("loading_begin_text"),
    SInt16("loading_end_text"),
    SInt16("checkpoint_begin_text"),
    SInt16("checkpoint_end_text"),
    dependency("checkpoint", "snd!"),
    BytearrayRaw("unknown", SIZE=96, VISIBLE=False),
    SIZE=120
    )

hudg_body = Struct("tagdata",
    messaging_parameters,
    reflexive("button_icons", button_icon, 18,
        "a button", "b button", "x button", "y button",
        "black button", "white button", "left trigger", "right trigger",
        "dpad up",  "dpad down", "dpad left", "dpad right", "start", "back",
        "left thumb button", "right thumb button", "left stick", "right stick"
        ),
    hud_help_text_color,
    dependency("hud_messages", "hmt "),
    objective_colors,
    waypoint_parameters,
    reflexive("waypoint_arrows", waypoint_arrow, 16, DYN_NAME_PATH='.name'),

    Pad(80),
    hud_globals,
    hud_damage_indicators,
    time_running_out,
    time_out,

    Pad(40),
    dependency("carnage_report_bitmap", "bitm"),
    misc_hud_crap,
    SIZE=1104
    )


def get():
    return hudg_def

hudg_def = TagDef("hudg",
    blam_header("hudg"),
    hudg_body,

    ext=".hud_globals", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from array import array
from reclaimer.constants import TYPE_CUBEMAP, CUBEMAP_PADDING, BITMAP_PADDING,\
     FORMAT_NAME_MAP, TYPE_NAME_MAP, FORMAT_P8_BUMP
from reclaimer.bitmaps.p8_palette import HALO_P8_PALETTE
from reclaimer.hek.defs.objs.tag import HekTag

try:
    import arbytmap as ab

    if not hasattr(ab, "FORMAT_P8_BUMP"):
        ab.FORMAT_P8_BUMP = "P8-BUMP"

        """ADD THE P8 FORMAT TO THE BITMAP CONVERTER"""
        ab.register_format(format_id=ab.FORMAT_P8_BUMP, depths=(8,8,8,8))
except (ImportError, AttributeError):
    ab = None


class BitmTag(HekTag):
    tex_infos = ()
    p8_palette = None

    def __init__(self, *args, **kwargs):
        HekTag.__init__(self, *args, **kwargs)
        self.p8_palette = HALO_P8_PALETTE

    def bitmap_count(self, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.size
        self.data.tagdata.bitmaps.size = new_value

    def bitmap_width(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].width
        self.data.tagdata.bitmaps.bitmaps_array[b_index].width = new_value

    def bitmap_height(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].height
        self.data.tagdata.bitmaps.bitmaps_array[b_index].height = new_value

    def bitmap_depth(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].depth
        self.data.tagdata.bitmaps.bitmaps_array[b_index].depth = new_value

    def bitmap_mipmaps_count(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].mipmaps
        self.data.tagdata.bitmaps.bitmaps_array[b_index].mipmaps = new_value

    def bitmap_type(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].type.data
        self.data.tagdata.bitmaps.bitmaps_array[b_index].type.data = new_value

    def bitmap_format(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].format.data
        self.data.tagdata.bitmaps.bitmaps_array[b_index].format.data = new_value

    def bitmap_width_height_depth(self, b_index=0, new_value=None):
        bitmap = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return(bitmap.width, bitmap.height, bitmap.depth)
        bitmap.width, bitmap.height, bitmap.depth = (
            new_value[0], new_value[1], new_value[2])

    def bitmap_flags(self, b_index=0, new_value=None):
        if new_value is None:
            return self.data.tagdata.bitmaps.bitmaps_array[b_index].flags
        self.data.tagdata.bitmaps.bitmaps_array[b_index].flags = new_value

    def bitmap_base_address(self, b_index=0, new_value=None):
        bitm = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return(bitm.base_address)
        bitm.base_address=new_value

    def bitmap_data_offset(self, b_index=0, new_value=None):
        bitm = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return(bitm.pixels_offset)
        bitm.pixels_offset=new_value

    def registration_point_x(self, b_index=0, new_value=None):
        bitm = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return bitm.registration_point_x
        bitm.registration_point_x = new_value

    def registration_point_y(self, b_index=0, new_value=None):
        bitm = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return bitm.registration_point_y
        bitm.registration_point_y = new_value

    def registration_point_xy(self, b_index=0, new_value=None):
        bitm = self.data.tagdata.bitmaps.bitmaps_array[b_index]
        if new_value is None:
            return(bitm.registration_point_x,
                   bitm.registration_point_y)
        bitm.registration_point_x, bitm.registration_point_y = new_value[0],\
                                                               new_value[1]

    @property
    def is_xbox_bitmap(self):
        # we only need to check the first bitmap
        if not self.bitmap_count(): return False
        return self.bitmap_base_address() == 1073751810

    def is_power_of_2_bitmap(self, b_index=0):
        if not self.bitmap_count(): return False
        return self.bitmap_flags(b_index).power_of_2_dim

    def is_compressed_bitmap(self, b_index=0):
        if not self.bitmap_count(): return False
        return self.bitmap_flags(b_index).compressed

    def swizzled(self, b_index=0, new_flag=None):
        if new_flag is None:
            if not self.bitmap_count(): return False
            return self.bitmap_flags(b_index).swizzled
        if not self.bitmap_count(): return
        self.bitmap_flags(b_index).swizzled = new_flag

    def color_plate_data_bytes_size(self, new_value=None):
        if new_value is None:
            return(self.data.tagdata.compressed_color_plate_data.size)
        self.data.tagdata.compressed_color_plate_data.size = new_value

    def pixel_data_bytes_size(self, new_value=None):
        if new_value is None:
            return self.data.tagdata.processed_pixel_data.size
        self.data.tagdata.processed_pixel_data.size = new_value

    def set_platform(self, saveasxbox):
        '''changes different things to set the platform to either PC or Xbox'''
        # read each of the bitmap blocks
        for b_index in range(self.bitmap_count()):
            bitmap = self.data.tagdata.bitmaps.bitmaps_array[b_index]

            bitmap.flags.prefer_low_detail = saveasxbox

            '''base_address is the ONLY discernable difference
            between a bitmap made by arsenic from a PC map, and
            a bitmap made by arsenic from an original XBOX map'''
            if saveasxbox:
                # change some miscellaneous variables
                bitmap.pixels = 18
                bitmap.bitmap_data_pointer = 0xFFFFFFFF
                bitmap.base_address = 1073751810
            else:
                bitmap.base_address = 0

        if not saveasxbox:
            return

        # if Xbox, reset these structure variable's all to 0
        # since xbox doesn't like them being non-zero
        tagdata = self.data.tagdata
        tagdata.compressed_color_plate_data.flags.data = 0
        tagdata.processed_pixel_data.flags.data = 0
        for i in (2,3):
            tagdata.compressed_color_plate_data[i] = 0
            tagdata.processed_pixel_data[i] = 0

        for i in (1,2):
            tagdata.sequences[i] = 0
            tagdata.bitmaps[i] = 0

        # swap the order of the cubemap faces
        # and mipmaps if saving to xbox format
        self.change_sub_bitmap_ordering(saveasxbox)

    def change_sub_bitmap_ordering(self, saveasxbox):
        '''Used to change the mipmap and cube face ordering.
        On pc all highest resolution faces are first, then
        the next highest resolution mipmap set. On xbox it's
        all of a face's mipmaps before any of the other faces.

        DO NOT UNDER ANY CIRCUMSTANCES CALL THIS FUNCTION
        IF PADDING HAS ALREADY BEEN ADDED TO A BITMAP'''

        raw_bitmap_data = self.data.tagdata.processed_pixel_data.data

        # loop over each of the bitmap blocks
        for b_index in range(self.bitmap_count()):
            if self.bitmap_type(b_index) == TYPE_CUBEMAP:
                mipmap_count = self.bitmap_mipmaps_count(b_index) + 1
                tex_block = raw_bitmap_data[b_index]

                # this will be used to copy values from
                template = tex_block.__copy__()

                # this is used to keep track of which index
                # we're placing the new pixel array into
                i = 0

                '''since we also want to swap the second and third
                cubemap faces we can do that easily like this.
                xbox has the second and third cubemap faces swapped
                with each other compared to pc. IDFKY'''
                for face in (0, 2, 1, 3, 4, 5):
                    for mip in range(0, mipmap_count*6, 6):
                        '''get the block we want from the original
                        layout and place it in its new position'''
                        if saveasxbox:
                            tex_block[i] = template[mip + face]
                        else:
                            tex_block[mip + face] = template[i]
                        i += 1

    def add_bitmap_padding(self, save_as_xbox):
        '''This function will create and apply padding to each of the
        bitmaps in the tag to make it XBOX compatible. This function will
        also add the number of bytes of padding to the internal offsets'''

        """The offset of each bitmap's pixel data needs to be increased by
        the padding of all the bitmaps before it. This variable will be
        used for knowing the total amount of padding before each bitmap.

        DO NOT RUN IF A BITMAP ALREADY HAS PADDING."""
        total_data_size = 0
        if ab is None:
            raise NotImplementedError(
                "Arbytmap is not loaded. Cannot add padding.")

        for i in range(self.bitmap_count()):
            sub_bitmap_count = 1
            if self.bitmap_type(i) == TYPE_CUBEMAP:
                sub_bitmap_count = 6

            pixel_data_block = self.data.tagdata.processed_pixel_data.data[i]

            # apply the offset to the tag
            self.bitmap_data_offset(i, total_data_size)

            if save_as_xbox or self.bitmap_format(i) == FORMAT_P8_BUMP:
                # calculate how much padding to add to the xbox bitmaps
                bitmap_pad, cubemap_pad = self.get_padding_size(i)

                # if this bitmap has padding on each of the sub-bitmaps
                if cubemap_pad:
                    mipmap_count = self.bitmap_mipmaps_count(i) + 1
                    for j in range(0, 6*(mipmap_count + 1), mipmap_count + 1):
                        pad = bytearray(cubemap_pad)
                        if isinstance(pixel_data_block[0], array):
                            pad = array('B', pad)
                        pixel_data_block.insert(j + mipmap_count, pad)

                # add the main padding to the end of the bitmap block
                pad = bytearray(bitmap_pad)
                if isinstance(pixel_data_block[0], array):
                    pad = array('B', pad)
                pixel_data_block.append(pad)

            # add the number of bytes this bitmap is to the
            # total bytes so far(multiple by sub-bitmap count)
            for pixel_data in pixel_data_block:
                if isinstance(pixel_data, array):
                    total_data_size += len(pixel_data) * pixel_data.itemsize
                else:
                    total_data_size += len(pixel_data)

        # update the total number of bytes of pixel data
        # in the tag by all the padding that was added
        self.pixel_data_bytes_size(total_data_size)

    def get_bitmap_size(self, b_index):
        '''Given a bitmap index, this function will
        calculate how many bytes the data takes up.
        THIS FUNCTION WILL NOT TAKE INTO ACCOUNT THE NUMBER OF SUB-BITMAPS'''
        if ab is None:
            raise NotImplementedError(
                "Arbytmap is not loaded. Cannot get bitmap size.")

        w, h, d, = self.bitmap_width_height_depth(b_index)
        fmt = FORMAT_NAME_MAP[self.bitmap_format(b_index)]

        bytes_count = 0
        for mipmap in range(self.bitmap_mipmaps_count(b_index) + 1):
            mw, mh, md = ab.get_mipmap_dimensions(w, h, d, mipmap)
            if fmt == ab.FORMAT_P8_BUMP:
                bytes_count += mw*mh*md
            else:
                bytes_count += ab.bitmap_io.get_pixel_bytes_size(fmt, mw, mh, md)

        return bytes_count

    def get_padding_size(self, b_index):
        bytes_count = self.get_bitmap_size(b_index)
        cubemap_pad = 0

        if self.bitmap_type(b_index) == TYPE_CUBEMAP:
            cubemap_pad = ((CUBEMAP_PADDING - (bytes_count % CUBEMAP_PADDING))
                           % CUBEMAP_PADDING)
            bytes_count = (bytes_count + cubemap_pad) * 6

        bitmap_pad = (BITMAP_PADDING -
                      (bytes_count%BITMAP_PADDING)) % BITMAP_PADDING

        return bitmap_pad, cubemap_pad

    def sanitize_mipmap_counts(self):
        '''Some original xbox bitmaps have fudged up mipmap counts
        and cause issues. This function will scan through all a
        bitmap's bitmaps and check that they fit within their
        calculated pixel data bounds. This is done by checking if a
        bitmap's calculated size is both within the side of the total
        pixel data and less than the next bitmap's pixel data start'''

        bad_bitmap_index_list = []
        bitmap_count = self.bitmap_count()

        for i in range(bitmap_count):
            # if this is the last bitmap
            if i + 1 == bitmap_count:
                # this is how many bytes of texture data there is total
                max_size = self.pixel_data_bytes_size()
            else:
                # this is the start of the next bitmap's pixel data
                max_size = self.bitmap_data_offset(i+1)

            while True:
                mipmap_count = self.bitmap_mipmaps_count(i)
                curr_size = self.get_bitmap_size(i) + self.bitmap_data_offset(i)

                if curr_size <= max_size:
                    break

                self.bitmap_mipmaps_count(i, mipmap_count - 1)

                # the mipmap count is zero and the bitmap still will
                # not fit within the space provided. Something's wrong
                if mipmap_count == 0:
                    bad_bitmap_index_list.append(i)
                    break

        return bad_bitmap_index_list

    def sanitize_bitmaps(self):
        if ab is None:
            raise NotImplementedError(
                "Arbytmap is not loaded. Cannot sanitize bitmaps.")
        tex_infos = self.tex_infos

        for i in range(self.bitmap_count()):
            format = FORMAT_NAME_MAP[self.bitmap_format(i)]
            flags = self.bitmap_flags(i)
            old_w, old_h, _ = self.bitmap_width_height_depth(i)

            reg_point_x, reg_point_y = self.registration_point_xy(i)
            texinfo = tex_infos[i]

            # set the flags to the new value
            flags.palletized = (format == ab.FORMAT_P8_BUMP)
            flags.compressed = (format in ab.COMPRESSED_FORMATS)

            self.bitmap_width_height_depth(
                i, (texinfo["width"], texinfo["height"], texinfo["depth"]))
            self.bitmap_mipmaps_count(i, texinfo["mipmap_count"])
            self.registration_point_xy(i, (texinfo["width"]*reg_point_x//old_w,
                                           texinfo["height"]*reg_point_y//old_h))

    def parse_bitmap_blocks(self):
        '''converts the raw pixel data into arrays of pixel
        data and replaces the raw data in the tag with them'''
        if ab is None:
            raise NotImplementedError(
                "Arbytmap is not loaded. Cannot parse bitmaps.")

        pixel_data = self.data.tagdata.processed_pixel_data
        rawdata = pixel_data.data

        tex_infos = self.tex_infos = []

        # this is the block that will hold all of the bitmap blocks
        root_tex_block = self.definition.subdefs['pixel_root'].build()

        is_xbox = self.is_xbox_bitmap
        get_mip_dims = ab.get_mipmap_dimensions
        bytes_to_array = ab.bitmap_io.bitmap_bytes_to_array

        # read the pixel data blocks for each bitmap
        for i in range(self.bitmap_count()):
            # since we need this information to read the bitmap we extract it
            mw, mh, md, = self.bitmap_width_height_depth(i)
            type         = self.bitmap_type(i)
            format       = FORMAT_NAME_MAP[self.bitmap_format(i)]
            mipmap_count = self.bitmap_mipmaps_count(i) + 1
            sub_bitmap_count = ab.SUB_BITMAP_COUNTS[TYPE_NAME_MAP[type]]

            # Get the offset of the pixel data for
            # this bitmap within the raw pixel data
            off = self.bitmap_data_offset(i)

            # this texture info is used in manipulating the texture data
            tex_infos.append(dict(
                width=mw, height=mh, depth=md, format=format,
                mipmap_count=mipmap_count-1, sub_bitmap_count=sub_bitmap_count,
                swizzled=self.swizzled(), texture_type=TYPE_NAME_MAP[type]))

            if format == ab.FORMAT_P8_BUMP:
                tex_infos[-1]["palette"] = [
                    self.p8_palette.p8_palette_32bit_packed]*mipmap_count

                # set it to packed since if we need to drop channels
                # then it needs to be unpacked with channels dropped
                tex_infos[-1]["palette_packed"] = True
                tex_infos[-1]["indexing_size"] = 8

            # this is the block that will hold each mipmap,
            # texture slice, and cube face of the bitmap
            root_tex_block.append()
            tex_block = root_tex_block[-1]

            # xbox bitmaps are stored all mip level faces first, then
            # the next mip level, whereas pc is the other way. Xbox
            # bitmaps also have padding between each mipmap and bitmap.
            dim0 = sub_bitmap_count if is_xbox else mipmap_count
            dim1 = mipmap_count if is_xbox else sub_bitmap_count
            for j in range(dim0):
                if not is_xbox: w, h, d = get_mip_dims(mw, mh, md, j)

                for k in range(dim1):
                    if is_xbox: w, h, d = get_mip_dims(mw, mh, md, k)

                    if format == ab.FORMAT_P8_BUMP:
                        pixel_count = w*h
                        tex_block.append(array('B', rawdata[off: off+pixel_count]))
                        off += pixel_count
                        continue

                    off = bytes_to_array(rawdata, off, tex_block, format, w, h, d)

                # skip the xbox alignment padding to get to the next texture
                if is_xbox:
                    tex_pad, sub_tex_pad = self.get_padding_size(i)
                    off += sub_tex_pad
                    if j + 1 == dim0:
                        off += tex_pad

        pixel_data.data = root_tex_block
        # now that we've successfully built the bitmap
        # blocks from the raw data we replace the raw data
        if is_xbox:
            # it's easier to work with bitmaps in one format so
            # we'll switch the mipmaps from XBOX to PC ordering
            self.change_sub_bitmap_ordering(False)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.antr import *
from ..common_descs import *

# As meta, it seems MOST arrays of anim_enum_desc have an extra extra on the end

animations_extended_desc = Struct("weapon_types",
    Pad(16),
    reflexive("animations", anim_enum_desc),
    SIZE=28,
    )

unit_weapon_desc = Struct("weapon",
    ascii_str32("name"),
    ascii_str32("grip_marker"),
    ascii_str32("hand_marker"),
    #Aiming screen bounds

    Pad(32),
    #pitch and yaw are saved in radians.
    float_rad("right_yaw_per_frame"),
    float_rad("left_yaw_per_frame"),
    SInt16("right_frame_count"),
    SInt16("left_frame_count"),

    float_rad("down_pitch_per_frame"),
    float_rad("up_pitch_per_frame"),
    SInt16("down_frame_count"),
    SInt16("up_frame_count"),

    reflexive("animations_extended", animations_extended_desc),
    reflexive("ik_points", ik_point_desc, 4, DYN_NAME_PATH=".marker"),
    reflexive("weapon_types", weapon_types_desc, DYN_NAME_PATH=".label"),
    SIZE=188,
    )

label_desc = Struct("label",
    ascii_str32("label"),
    SIZE=32
    )

unknown_unit_desc = Struct("unknown_unit_desc",
    SInt32("unknown0"),
    SInt32("unknown1"),
    ascii_str32("label"),
    reflexive("labels", label_desc),
    reflexive("animations", anim_enum_desc),
    SIZE=64
    )

unit_desc = Struct("unit",
    ascii_str32("label"),
    #pitch and yaw are saved in radians.

    #Looking screen bounds
    float_rad("right_yaw_per_frame"),
    float_rad("left_yaw_per_frame"),
    SInt16("right_frame_count"),
    SInt16("left_frame_count"),

    float_rad("down_pitch_per_frame"),
    float_rad("up_pitch_per_frame"),
    SInt16("down_frame_count"),
    SInt16("up_frame_count"),

    Pad(8),
    reflexive("animations", anim_enum_desc),
    reflexive("ik_points", ik_point_desc, 4, DYN_NAME_PATH=".marker"),
    reflexive("weapons", unit_weapon_desc, DYN_NAME_PATH=".name"),
    reflexive("unknown", unknown_unit_desc),
    SIZE=128,
    )

seat_desc = Struct("seat",
    ascii_str32("label"),
    Pad(16),
    reflexive("animations", anim_enum_desc),
    SIZE=60
    )

vehicle_desc = Struct("vehicle",
    #pitch and yaw are saved in radians.

    #Steering screen bounds
    float_rad("right_yaw_per_frame"),
    float_rad("left_yaw_per_frame"),
    SInt16("right_frame_count"),
    SInt16("left_frame_count"),

    float_rad("down_pitch_per_frame"),
    float_rad("up_pitch_per_frame"),
    SInt16("down_frame_count"),
    SInt16("up_frame_count"),

    Pad(56),
    reflexive("seats", seat_desc),
    reflexive("animations", anim_enum_desc, 8,
        'steering','roll','throttle','velocity',
        'braking','ground-speed','occupied','unoccupied'),
    reflexive("suspension_animations", suspension_desc, 8),
    SIZE=116,
    )

effect_reference_desc = Struct("effect_reference",
    dependency_stubbs('effect', ("snd!", "effe")),
    SIZE=20,
    )

animation_desc = Struct("animation",
    ascii_str32("name"),
    SEnum16("type", *anim_types),
    SInt16("frame_count"),
    SInt16("frame_size"),
    SEnum16("frame_info_type", *anim_frame_info_types),
    SInt32("node_list_checksum"),
    SInt16("node_count"),
    SInt16("loop_frame_index"),

    Float("weight"),
    SInt16("key_frame_index"),
    SInt16("second_key_frame_index"),
    Pad(8),

    dyn_senum16("next_animation", DYN_NAME_PATH="..[DYN_I].name"),
    Bool16("flags",
        "compressed_data",
        "world_relative",
        {NAME:"pal", GUI_NAME:"25Hz(PAL)"},
        ),
    dyn_senum16("sound",
        DYN_NAME_PATH="tagdata.effect_references." +
        "effect_references_array[DYN_I].effect.filepath"),
    SInt16("sound_frame_index"),
    SInt8("left_foot_frame_index"),
    SInt8("right_foot_frame_index"),
    FlSInt16("first_permutation_index", VISIBLE=False,
        TOOLTIP="The index of the first animation in the permutation chain."),
    FlFloat("chance_to_play", VISIBLE=False,
        MIN=0.0, MAX=1.0, SIDETIP="[0,1]",
        TOOLTIP=("Seems to be the chance range to select this permutation.\n"
                 "Random number in the range [0,1] is rolled. The permutation\n"
                 "chain is looped until the number is higher than or equal\n"
                 "to that permutations chance to play. This chance to play\n"
                 "is likely influenced by the animations 'weight' field.\n"
                 "All permutation chains should have the last one end with\n"
                 "a chance to play of 1.0")),

    rawdata_ref("frame_info", max_size=32768),
    UInt32("trans_flags0", EDITABLE=False),
    UInt32("trans_flags1", EDITABLE=False),
    Pad(8),
    UInt32("rot_flags0", EDITABLE=False),
    UInt32("rot_flags1", EDITABLE=False),
    Pad(8),
    UInt32("scale_flags0", EDITABLE=False),
    UInt32("scale_flags1", EDITABLE=False),
    Pad(4),
    SInt32("offset_to_compressed_data", EDITABLE=False),
    rawdata_ref("default_data", max_size=16384),
    rawdata_ref("frame_data", max_size=1048576),
    SIZE=188,
    )

antr_body = Struct("tagdata",
    reflexive("objects",  object_desc),
    reflexive("units",    unit_desc, DYN_NAME_PATH=".label"),
    reflexive("weapons",  weapon_desc),
    reflexive("vehicles", vehicle_desc),
    reflexive("devices",  device_desc),
    reflexive("unit_damages", anim_enum_desc),
    reflexive("fp_animations", fp_animation_desc),

    reflexive("effect_references", effect_reference_desc,
        DYN_NAME_PATH=".effect.filepath"),
    Float("limp_body_node_radius"),
    Bool16("flags",
        "compress_all_animations",
        "force_idle_compression",
        ),
    Pad(2),
    reflexive("nodes", nodes_desc, DYN_NAME_PATH=".name"),
    reflexive("animations", animation_desc, DYN_NAME_PATH=".name"),
    SIZE=128,
    )


def get():
    return antr_def

antr_def = TagDef("antr",
    blam_header_stubbs('antr', 5),  # increment to differentiate it from halo antr
    antr_body,

    ext=".model_animations", endian=">", tag_cls=AntrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

def get():
    return srud_def

srud_def = TagDef("srud",
    blam_header('srud'),
    Struct('tagdata',
        BytesRaw("unknown", SIZE=5120),
        SIZE=5120
        ),

    ext=".shadowrun_unit", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

transform_states_comment = """
The transform out tag determines the cause of the transformation(such as low health),
the animation to play on the attacked unit, and what attachments to add, if any.
The transformation in tag determines the possible actors to transform into.
"""

transform = Struct("transform",
    Bool16("flags",
        'scripted_only',
        ),
    Pad(2),
    ascii_str32("transform_name"),
    QStruct("selection_chances",
        Float("easy"), Float("normal"), Float("hard"), Float("imposs"),
        ORIENT="h"
        ),
    Struct("transform_stages",
        dependency_os("transform_out", 'avto'),
        Pad(4),
        dependency_os("transform_in", 'avti'),
        COMMENT=transform_states_comment
        ),
    SIZE=116
    )

actor_variant_transform = Struct("actor_variant_transform",
    dependency_os("actor_variant", 'actv'),
    reflexive("transforms",
        transform, 32, DYN_NAME_PATH='.transform_name'),
    SIZE=52
    )

avtc_body = Struct("tagdata",
    reflexive("actor_variant_transforms",
        actor_variant_transform, 32, DYN_NAME_PATH='.actor_variant.filepath'),
    SIZE=36
    )

def get():
    return avtc_def

avtc_def = TagDef("avtc",
    blam_header_os('avtc', 1),
    avtc_body,

    ext=".actor_variant_transform_collection", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.obje import ObjeTag
from reclaimer.hek.defs.objs.devi import DeviTag

class LifiTag(DeviTag, ObjeTag):

    def calc_internal_data(self):
        ObjeTag.calc_internal_data(self)
        DeviTag.calc_internal_data(self)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Standardizing
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


effg_unknown_unknown_2 = Struct("unknown_2",
    SInt32("unknown_0"),
    SInt32("unknown_1"),
    Pad(4),
    SInt32("unknown_3"),
    ENDIAN=">", SIZE=16
    )


effg_unknown = Struct("unknown",
    SInt32("unknown_0"),
    SInt32("unknown_1"),
    h3_reflexive("unknown_2", effg_unknown_unknown_2),
    ENDIAN=">", SIZE=20
    )


effg_body = Struct("tagdata",
    h3_reflexive("unknown", effg_unknown),
    ENDIAN=">", SIZE=12
    )


def get():
    return effg_def

effg_def = TagDef("effg",
    h3_blam_header('effg'),
    effg_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["effg"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path

from reclaimer.h2.util import split_raw_pointer
from reclaimer.sounds import constants
from reclaimer.sounds import util
from reclaimer.sounds.blam_sound_bank import BlamSoundBank, BlamSoundPitchRange
from reclaimer.sounds.blam_sound_permutation import BlamSoundPermutation
from reclaimer.sounds.blam_sound_samples import BlamSoundSamples


__all__ = ("extract_h1_sounds", "extract_h2_sounds", )


def extract_h1_sounds(tagdata, tag_path, **kw):
    do_write_wav = kw.get('write_wav', True)
    overwrite = kw.get('overwrite', True)
    decompress = kw.get('decode_adpcm', True)
    pcm_is_big_endian = kw.get('byteswap_pcm_samples', False)
    tagpath_base = Path(kw['out_dir']).joinpath(Path(tag_path).with_suffix(""))

    encoding = tagdata.encoding.data
    channels = constants.channel_counts.get(encoding, 1)
    sample_rate = constants.halo_1_sample_rates.get(
        tagdata.sample_rate.data, 0)
    compression = constants.halo_1_compressions.get(
        tagdata.compression.data, constants.COMPRESSION_UNKNOWN)

    # make a BlamSoundBank to store all our sound data
    sound_bank = BlamSoundBank()
    sound_bank.split_into_smaller_chunks = tagdata.flags.split_long_sound_into_permutations
    sound_bank.split_to_adpcm_blocksize = tagdata.flags.fit_to_adpcm_blocksize
    sound_bank.sample_rate = sample_rate
    sound_bank.compression = compression
    sound_bank.encoding = encoding

    same_pr_names = {}
    for i, pr in enumerate(tagdata.pitch_ranges.STEPTREE):
        # determine a usable name for this pitch range
        pr_name = util.BAD_PATH_CHAR_REMOVAL.sub("_", pr.name.strip())
        if not pr_name:
            pr_name = str(i)

        same_pr_names[pr_name] = same_pr_names.get(pr_name, -1) + 1
        if same_pr_names[pr_name]:
            pr_name += "_%s" % same_pr_names[pr_name]

        # make a BlamPitchRange to store all the permutations
        sound_bank.pitch_ranges[pr_name] = BlamSoundPitchRange()

        same_names_permlists = {}
        unchecked_perms = set(range(len(pr.permutations.STEPTREE)))

        perm_indices = list(
            range(min(pr.actual_permutation_count, len(unchecked_perms))))
        if not perm_indices:
            perm_indices = set(unchecked_perms)

        playback_speed = 1.0
        if pr.natural_pitch > 0:
            playback_speed = 1 / pr.natural_pitch

        while perm_indices:
            # loop over all of the actual permutation indices and combine
            # the permutations they point to into a list with a shared name.
            # we do this so we can combine the permutations together into one.
            for j in perm_indices:
                perm = pr.permutations.STEPTREE[j]
                compression = perm.compression.enum_name
                name = perm.name if perm.name else str(j)
                permlist = same_names_permlists.get(name, [])
                same_names_permlists[name] = permlist

                while j in unchecked_perms:
                    perm = pr.permutations.STEPTREE[j]
                    if compression != perm.compression.enum_name:
                        # cant combine when compression is different
                        break

                    unchecked_perms.remove(j)
                    permlist.append(perm)

                    j = perm.next_permutation_index
            perm_indices = set(unchecked_perms)

        for name, permlist in same_names_permlists.items():
            blam_permutation = BlamSoundPermutation(
                sample_rate=sample_rate, encoding=encoding)
            sound_bank.pitch_ranges[pr_name].permutations[name] = blam_permutation

            for perm in permlist:
                sample_data = perm.samples.data
                if perm.compression.enum_name == "ogg":
                    # not actually a sample count. fix struct field name
                    compression = constants.COMPRESSION_OGG
                    sample_count = perm.ogg_sample_count // 2
                elif perm.compression.enum_name == "none":
                    compression = (constants.COMPRESSION_PCM_16_BE
                                   if pcm_is_big_endian else
                                   constants.COMPRESSION_PCM_16_LE)
                    sample_count = len(sample_data) // 2
                elif perm.compression.enum_name == "xbox_adpcm":
                    compression = constants.COMPRESSION_XBOX_ADPCM
                    sample_count = (
                        (constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2) *
                        (len(sample_data) // constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE))
                elif perm.compression.enum_name == "ima_adpcm":
                    compression = constants.COMPRESSION_IMA_ADPCM
                    sample_count = (
                        (constants.IMA_ADPCM_DECOMPRESSED_BLOCKSIZE // 2) *
                        (len(sample_data) // constants.IMA_ADPCM_COMPRESSED_BLOCKSIZE))
                else:
                    print("Unknown audio compression type:", perm.compression.data)
                    continue

                sample_count = sample_count // channels
                blam_permutation.processed_samples.append(
                    BlamSoundSamples(
                        sample_data, sample_count, compression,
                        int(round(sample_rate * playback_speed)),
                        encoding, perm.mouth_data.data)
                    )

    if do_write_wav:
        sound_bank.export_to_directory(
            tagpath_base, overwrite, False, decompress)
    else:
        return sound_bank


def get_sound_name(import_names, index):
    if index < 0 or index >= len(import_names):
        return ""
    return import_names[index].string


def extract_h2_sounds(tagdata, tag_path, **kw):
    # TODO: Make this multiply the sample rate by the natural pitch

    halo_map = kw.get('halo_map')
    if not halo_map:
        print("Cannot run this function on tags.")
        return

    do_write_wav = kw.get('write_wav', True)
    overwrite = kw.get('overwrite', True)
    decompress = kw.get('decode_adpcm', True)
    tagpath_base = Path(kw['out_dir']).joinpath(Path(tag_path).with_suffix(""))

    ugh__meta = halo_map.ugh__meta
    if ugh__meta is None:
        return "    No sound_cache_file_gestalt to read sounds from."

    import_names = ugh__meta.import_names.STEPTREE
    pitch_ranges = ugh__meta.pitch_ranges.STEPTREE
    permutations = ugh__meta.permutations.STEPTREE
    perm_chunks  = ugh__meta.permutation_chunks.STEPTREE

    same_pr_names = {}

    encoding = tagdata.encoding.data
    sample_rate = constants.halo_2_sample_rates.get(
        tagdata.sample_rate.data, 0)
    compression = constants.halo_2_compressions.get(
        tagdata.compression.data, constants.COMPRESSION_UNKNOWN)

    if encoding == constants.ENCODING_CODEC:
        compression = constants.COMPRESSION_UNKNOWN

    if tagdata.encoding.enum_name == "codec":
        pass # return "    CANNOT YET EXTRACT THIS FORMAT."
        '''
        The codec format seems to be encoded with wmaudio2.

        Bytes:
            0-31
                Two nearly identical GUIDs, with the second one specifying in
                its first 2 bytes that the audio data is encoded with wmaudio2.
            32-39
                Unknown, but always seems to be 01 00 00 00  00 00 00 00
            40-43
                Number of blocks of data(same value found in the header below)
            44-55
                Unknown, but I think it always seems to be the same.
            56-87
                wav_format header struct. sig is "ZYU\x00" instead of "fmt ",
                length is 28 instead of 20, and fmt is(always?) 0x0161,
                which is the format code for wmaudio2.
                channels, sample_rate, byte_rate, block_align, and
                bits_per_sample all seem to be set properly. the value
                for block_align is the same as bytes 40-43
            88-91
                Unknown. Might be something to specify the data length?

            Everything after this appears to be audio data.
        '''

    # get the pitch range indices to iterate over
    pr_index = tagdata.pitch_range_index
    pr_count = min(tagdata.pitch_range_count, len(pitch_ranges) - pr_index)
    if pr_index < 0 or pr_count < 0:
        # nothing to extract
        return

    # make a BlamSoundBank to store all our sound data
    sound_bank = BlamSoundBank()
    sound_bank.split_into_smaller_chunks = tagdata.flags.split_long_sound_into_permutations
    sound_bank.split_to_adpcm_blocksize = tagdata.flags.fit_to_adpcm_blocksize
    sound_bank.sample_rate = sample_rate
    sound_bank.compression = compression
    sound_bank.encoding = encoding

    for i in range(pr_index, pr_index + pr_count):
        pr = pitch_ranges[i]
        pr_name = get_sound_name(import_names, pr.import_name_index)
        pr_name = util.BAD_PATH_CHAR_REMOVAL.sub(
            "_", pr_name).strip().replace('|', ' ')
        if not pr_name:
            pr_name = str(i)

        same_pr_names[pr_name] = same_pr_names.get(pr_name, -1) + 1
        if same_pr_names[pr_name]:
            pr_name += "_%s" % same_pr_names[pr_name]

        # make a BlamPitchRange to store all the permutations
        sound_bank.pitch_ranges[pr_name] = BlamSoundPitchRange()

        # get the permutation indices to iterate over
        perm_index = pr.first_permutation
        perm_count = min(pr.permutation_count, len(permutations) - perm_index)
        if perm_index < 0 or perm_count < 0:
            continue

        for j in range(perm_index, perm_index + perm_count):
            perm = permutations[j]
            perm_name = get_sound_name(
                import_names, perm.import_name_index).replace('|', ' ')
            if not perm_name: perm_name = str(j)

            blam_permutation = BlamSoundPermutation(
                sample_rate=sample_rate, encoding=encoding)
            sound_bank.pitch_ranges[pr_name].permutations[perm_name] = blam_permutation

            # get the chunk indices to iterate over
            chunk_index = perm.first_chunk
            chunk_count = min(perm.chunk_count, len(perm_chunks) - chunk_index)
            if chunk_index < 0 or chunk_count < 0:
                continue

            for k in range(chunk_index, chunk_index + chunk_count):
                ptr, map_name = split_raw_pointer(perm_chunks[k].pointer)
                this_map = halo_map
                if map_name != "local":
                    this_map = halo_map.maps.get(map_name)

                if this_map is None or this_map.map_data is None:
                    continue

                this_map.map_data.seek(ptr)
                blam_permutation.processed_samples.append(
                    BlamSoundSamples(
                        this_map.map_data.read(perm_chunks[k].size),
                        0, compression, sample_rate, encoding)
                    )

    if do_write_wav:
        sound_bank.export_to_directory(
            tagpath_base, overwrite, False, decompress)
    else:
        return sound_bank



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from reclaimer.hek.defs.objs.tag import HekTag

class ObjeTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        full_class_name = self.data.blam_header.tag_class.enum_name

        self.ext = '.' + full_class_name
        self.filepath = os.path.splitext(str(self.filepath))[0] + self.ext

        object_type = self.data.tagdata.obje_attrs.object_type
        if full_class_name == "object":
            object_type.data = -1
        elif full_class_name == "biped":
            object_type.data = 0
        elif full_class_name == "vehicle":
            object_type.data = 1
        elif full_class_name == "weapon":
            object_type.data = 2
        elif full_class_name == "equipment":
            object_type.data = 3
        elif full_class_name == "garbage":
            object_type.data = 4
        elif full_class_name == "projectile":
            object_type.data = 5
        elif full_class_name == "scenery":
            object_type.data = 6
        elif full_class_name == "device_machine":
            object_type.data = 7
        elif full_class_name == "device_control":
            object_type.data = 8
        elif full_class_name == "device_light_fixture":
            object_type.data = 9
        elif full_class_name == "placeholder":
            object_type.data = 10
        elif full_class_name == "sound_scenery":
            object_type.data = 11
        else:
            raise ValueError("Unknown object type '%s'" % full_class_name)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#
from pathlib import Path

from supyr_struct.defs.constants import DEFAULT
from supyr_struct.tag import Tag

from reclaimer.util import calc_halo_crc32

class HekTag(Tag):
    def __init__(self, **kwargs):
        self.calc_pointers = False
        Tag.__init__(self, **kwargs)

    def serialize(self, **kwargs):
        '''
        Overload of the supyr serialization function that retroactively adds
        a CRC to the tag.
        '''
        head = self.data.blam_header
        filepath = kwargs.get('filepath', self.filepath)
        buffer = kwargs.get('buffer', None)

        # Run the normal serialization.
        result = Tag.serialize(self, **kwargs)

        # If there is neither a buffer or filepath just return the result.
        if (buffer is None) and (not filepath):
            return result

        # Prefer to use the buffer as that is how Tag.serialize does it.
        f = buffer
        if buffer is None:
            f = Path(filepath).open('rb+')

        # Calculate the crc from after the header to the end.
        crc = calc_halo_crc32(f, offset=head.get_desc('SIZE'))
        # Write the crc to the offset of the checksum value in the header.
        # The way we retrieve this offset from supyr is insane.
        attr_index = head.get_desc('NAME_MAP')['checksum']
        f.seek(head.get_desc('ATTR_OFFS')[attr_index])
        f.write(crc.to_bytes(4, byteorder='big', signed=False))
        # Flush the stream.
        f.flush()
        # Only close if it is a file. Because the only case where we own
        # this buffer is if there was no buffer kwarg.
        if not buffer:
            f.close()

        # Update the tag object so it won't have to be deserialized again.
        head.checksum = crc
        return result

    def calc_internal_data(self):
        # recalculate the header data
        head = self.data.blam_header

        head.tag_class.data = head.tag_class.get_desc(DEFAULT)
        head.flags.edited_with_mozz = True
        head.header_size = head.get_desc(DEFAULT, 'header_size')
        head.version = head.get_desc(DEFAULT, 'version')
        head.integrity0 = head.get_desc(DEFAULT, 'integrity0')
        head.integrity1 = head.get_desc(DEFAULT, 'integrity1')
        head.engine_id.data = head.engine_id.get_desc(DEFAULT)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

radiosity_comment = """RADIOSITY/LIGHTMAPPING
The simple parameterisation flag is used for shaders that are on surfaces with more curved
shapes like grass hills. It avoids lightmap uv problems by using the uvs from the bsp geometry.
The ignore normals flag helps with shading bugs on double sided polygons by making the
light independent of normals.(Suggested use: tree leaves).

The detail level controls the relative quality of the lightmaps/lightmap uvs for this shader."""

lighting_comment = """LIGHT
Light power controls the brightness and reach of the light emitted by this shader
during lightmap rendering.
"""

material_type_comment = """MATERIAL TYPE
The material type is used to determine what material effects should be used for impacts on
BSP geometry that uses this shader."""

shdr_attrs = Struct("shdr_attrs",
    Bool16("radiosity_flags",
        { NAME: "simple_parameterization", GUI_NAME: "simple parameterization(lightmap fix)" },
        "ignore_normals",
        "transparent_lit",
        COMMENT=radiosity_comment
        ),
    SEnum16("radiosity_detail_level" ,
        "high",
        "medium",
        "low",
        "turd",
        ),
    Float("radiosity_light_power", COMMENT=lighting_comment),
    QStruct("radiosity_light_color", INCLUDE=rgb_float),
    QStruct("radiosity_tint_color",  INCLUDE=rgb_float),

    Pad(2),
    SEnum16("material_type", *materials_list, COMMENT=material_type_comment),
    # THIS FIELD IS OFTEN INCORRECT ON STOCK TAGS.
    FlSEnum16("shader_type",
        *((shader_types[i], i - 1) for i in
          range(len(shader_types))),
        VISIBLE=False, DEFAULT=-1
        ),
    Pad(2),
    SIZE=40
    )

shader_body = Struct("tagdata",
    shdr_attrs,
    SIZE=40
    )

def get():
    return shdr_def

shdr_def = TagDef("shdr",
    blam_header('shdr'),
    shader_body,

    ext=".shader", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import math

from array import array
from copy import deepcopy
from struct import pack_into, unpack_from
from types import MethodType

from reclaimer.animation.jma import JmaRootNodeState, JmaNodeState
from reclaimer.animation.structs import compressed_frames_def
from reclaimer.util import compression
from reclaimer.util import matrices

__all__ = ("serialize_frame_info", "deserialize_frame_info",
           "serialize_default_data", "deserialize_default_data",
           "serialize_frame_data", "deserialize_frame_data",
           "deserialize_compressed_frame_data",
           "serialize_compressed_frame_data",
           "get_anim_flags", "get_keyframe_index_of_frame",)


def serialize_frame_info(jma_anim, endian=">"):
    frame_ct = jma_anim.frame_count - 1
    data = bytearray(jma_anim.root_node_info_frame_size * frame_ct)

    pack_2_float_into = MethodType(pack_into, endian + "2f")
    pack_3_float_into = MethodType(pack_into, endian + "3f")
    pack_4_float_into = MethodType(pack_into, endian + "4f")
    frame_info_node_size = jma_anim.root_node_info_frame_size

    i = 0
    # write to the data
    if jma_anim.has_dz:
        for info in jma_anim.root_node_info[: frame_ct]:
            pack_4_float_into(
                data, i, info.dx / 100, info.dy / 100, info.dz / 100, info.dyaw)
            i += frame_info_node_size

    elif jma_anim.has_dyaw:
        for info in jma_anim.root_node_info[: frame_ct]:
            pack_3_float_into(
                data, i, info.dx / 100, info.dy / 100, info.dyaw)
            i += frame_info_node_size

    elif jma_anim.has_dxdy:
        for info in jma_anim.root_node_info[: frame_ct]:
            pack_2_float_into(
                data, i, info.dx / 100, info.dy / 100)
            i += frame_info_node_size

    return data


def deserialize_frame_info(anim, include_extra_base_frame=False, endian=">"):
    i = 0
    dx = dy = dz = dyaw = x = y = z = yaw = 0.0

    root_node_info = [JmaRootNodeState() for i in range(anim.frame_count)]
    frame_info = anim.frame_info.data

    # write to the data
    if "dz" in anim.frame_info_type.enum_name:
        unpack = MethodType(unpack_from, endian + "4f")
        for f in range(anim.frame_count):
            dx, dy, dz, dyaw = unpack(frame_info, i)
            dx *= 100; dy *= 100; dz *= 100

            info = root_node_info[f]
            info.dx = dx; info.dy = dy; info.dz = dz; info.dyaw = dyaw
            info.x  = x;  info.y  = y;  info.z  = z;  info.yaw  = yaw

            x += dx; y += dy; z += dz; yaw += dyaw
            i += 16

    elif "dyaw" in anim.frame_info_type.enum_name:
        unpack = MethodType(unpack_from, endian + "3f")
        for f in range(anim.frame_count):
            dx, dy, dyaw = unpack(frame_info, i)
            dx *= 100; dy *= 100

            info = root_node_info[f]
            info.dx = dx; info.dy = dy; info.dyaw = dyaw
            info.x  = x;  info.y  = y;  info.yaw  = yaw

            x += dx; y += dy; yaw += dyaw
            i += 12

    elif "dx" in anim.frame_info_type.enum_name:
        unpack = MethodType(unpack_from, endian + "2f")
        for f in range(anim.frame_count):
            dx, dy = unpack(frame_info, i)
            dx *= 100; dy *= 100

            info = root_node_info[f]
            info.dx = dx; info.dy = dy
            info.x  = x;  info.y  = y

            x += dx; y += dy
            i += 8

    if include_extra_base_frame and root_node_info:
        # duplicate the last frame and apply the change
        # that frame to the total change at that frame.
        last_root_node_info = deepcopy(root_node_info[-1])
        last_root_node_info.x += last_root_node_info.dx
        last_root_node_info.y += last_root_node_info.dy
        last_root_node_info.z += last_root_node_info.dz
        last_root_node_info.yaw += last_root_node_info.dyaw

        # no delta on last frame. zero it out
        last_root_node_info.dx = 0.0
        last_root_node_info.dy = 0.0
        last_root_node_info.dz = 0.0
        last_root_node_info.dyaw = 0.0

        root_node_info.append(last_root_node_info)

    return root_node_info


def serialize_default_data(jma_anim, endian=">"):
    data = bytearray(jma_anim.default_data_size)

    rot_flags   = jma_anim.rot_flags
    trans_flags = jma_anim.trans_flags
    scale_flags = jma_anim.scale_flags

    pack_1_float_into = MethodType(pack_into, endian +  "f")
    pack_3_float_into = MethodType(pack_into, endian + "3f")
    pack_4_int16_into = MethodType(pack_into, endian + "4h")

    sqrt = math.sqrt

    i = 0
    # write to the data
    def_frames = jma_anim.frames[0]
    for n in range(jma_anim.node_count):
        node_state = def_frames[n]
        if not rot_flags[n]:
            # components are ones-signed
            qi = node_state.rot_i
            qj = node_state.rot_j
            qk = node_state.rot_k
            qw = node_state.rot_w
            nmag = qi**2 + qj**2 + qk**2 + qw**2
            if nmag:
                nmag = 32767.5 / sqrt(nmag)
                qi = int(qi*nmag)
                qj = int(qj*nmag)
                qk = int(qk*nmag)
                qw = int(qw*nmag)
            else:
                qi = qj = qk = 0
                qw = 32767

            pack_4_int16_into(data, i, qi, qj, qk, qw)
            i += 8

        if not trans_flags[n]:
            pack_3_float_into(data, i,
                node_state.pos_x / 100,
                node_state.pos_y / 100,
                node_state.pos_z / 100)
            i += 12

        if not scale_flags[n]:
            pack_1_float_into(data, i, node_state.scale)
            i += 4

    return data


def deserialize_default_data(anim, endian=">"):
    return _deserialize_frame_data(anim, True, (), endian)[0]


def serialize_frame_data(jma_anim, endian=">"):
    data = bytearray(jma_anim.frame_data_frame_size *
                     (jma_anim.frame_count - 1))

    rot_flags   = jma_anim.rot_flags
    trans_flags = jma_anim.trans_flags
    scale_flags = jma_anim.scale_flags

    pack_1_float_into = MethodType(pack_into, endian +  "f")
    pack_3_float_into = MethodType(pack_into, endian + "3f")
    pack_4_int16_into = MethodType(pack_into, endian + "4h")

    is_overlay = jma_anim.anim_type == "overlay"

    sqrt = math.sqrt

    i = 0
    for f in range(jma_anim.frame_count):
        if not is_overlay and f + 1 == jma_anim.frame_count:
            # skip the last frame for non-overlays
            break
        elif f == 0 and is_overlay:
            # skip the first frame for overlays
            continue

        # write to the data
        for n in range(jma_anim.node_count):
            node_state = jma_anim.frames[f][n]

            if rot_flags[n]:
                # components are ones-signed
                qi = node_state.rot_i
                qj = node_state.rot_j
                qk = node_state.rot_k
                qw = node_state.rot_w
                nmag = qi**2 + qj**2 + qk**2 + qw**2
                if nmag:
                    nmag = 32767.5 / sqrt(nmag)
                    qi = int(qi*nmag)
                    qj = int(qj*nmag)
                    qk = int(qk*nmag)
                    qw = int(qw*nmag)
                else:
                    qi = qj = qk = 0
                    qw = 32767

                pack_4_int16_into(data, i, qi, qj, qk, qw)
                i += 8

            if trans_flags[n]:
                pack_3_float_into(data, i,
                    node_state.pos_x / 100,
                    node_state.pos_y / 100,
                    node_state.pos_z / 100)
                i += 12

            if scale_flags[n]:
                pack_1_float_into(data, i, node_state.scale)
                i += 4

    return data


def deserialize_frame_data(anim, def_node_states=None,
                           include_extra_base_frame=False, endian=">"):
    if def_node_states is None:
        def_node_states = deserialize_default_data(anim, endian)

    frame_data = _deserialize_frame_data(anim, False, def_node_states, endian)

    if not include_extra_base_frame:
        pass
    if anim.type.enum_name != "overlay":
        # duplicate the first frame to the last frame for non-overlays
        frame_data.append(deepcopy(frame_data[0]))
    else:
        # overlay animations start with frame 0 being
        # in the same state as the default node states
        frame_data.insert(0, def_node_states)

    return frame_data


def _deserialize_frame_data(anim, get_default_data, def_node_states, endian):
    unpack_trans = MethodType(unpack_from, endian + "3f")
    unpack_ijkw  = MethodType(unpack_from, endian + "4h")
    unpack_float = MethodType(unpack_from, endian +  "f")
    sqrt = math.sqrt

    rot_flags, trans_flags, scale_flags = get_anim_flags(anim)

    if get_default_data:
        store = False
        stored_frame_count = 1
        data = anim.default_data.data
    else:
        store = True
        stored_frame_count = anim.frame_count
        data = anim.frame_data.data

    all_node_states = [[JmaNodeState() for n in range(anim.node_count)]
                       for f in range(stored_frame_count)]

    if get_default_data:
        def_node_states = all_node_states[0]

    assert len(def_node_states) == anim.node_count

    i = 0
    for f in range(stored_frame_count):
        node_states = all_node_states[f]

        for n in range(anim.node_count):
            def_node_state = def_node_states[n]
            state = node_states[n]

            qi = qj = qk = x = y = z = 0.0
            qw = scale = 1.0
            if rot_flags[n] == store:
                qi, qj, qk, qw = unpack_ijkw(data, i)
                i += 8

                rot_len = qi**2 + qj**2 + qk**2 + qw**2
                if rot_len:
                    rot_len = 1 / sqrt(rot_len)
                    qi *= rot_len
                    qj *= rot_len
                    qk *= rot_len
                    qw *= rot_len
                else:
                    qi = qj = qk = 0.0
                    qw = 1.0
            else:
                qi = def_node_state.rot_i
                qj = def_node_state.rot_j
                qk = def_node_state.rot_k
                qw = def_node_state.rot_w

            if trans_flags[n] == store:
                x, y, z = unpack_trans(data, i)
                i += 12

                x *= 100
                y *= 100
                z *= 100
            else:
                x = def_node_state.pos_x
                y = def_node_state.pos_y
                z = def_node_state.pos_z

            if scale_flags[n] == store:
                scale = unpack_float(data, i)[0]
                i += 4
            else:
                scale = def_node_state.scale

            state.pos_x = x; state.pos_y = y; state.pos_z = z
            state.rot_i = qi; state.rot_j = qj
            state.rot_k = qk; state.rot_w = qw
            state.scale = scale

    return all_node_states


def serialize_compressed_frame_data(jma_anim):
    # make a compressed block to store the data for serialization
    comp_anim_block = compressed_frames_def.build()

    # local references for all these things to make access faster
    rot_flags   = jma_anim.rot_flags
    trans_flags = jma_anim.trans_flags
    scale_flags = jma_anim.scale_flags

    rot_keyframes   = comp_anim_block.rotation.keyframes
    trans_keyframes = comp_anim_block.translation.keyframes
    scale_keyframes = comp_anim_block.scale.keyframes

    rot_def_data   = comp_anim_block.rotation.default_data
    trans_def_data = comp_anim_block.translation.default_data
    scale_def_data = comp_anim_block.scale.default_data

    rot_keyframe_data   = comp_anim_block.rotation.keyframe_data
    trans_keyframe_data = comp_anim_block.translation.keyframe_data
    scale_keyframe_data = comp_anim_block.scale.keyframe_data

    rot_keyframes_by_nodes   = [[] for i in range(jma_anim.node_count)]
    trans_keyframes_by_nodes = [[] for i in range(jma_anim.node_count)]
    scale_keyframes_by_nodes = [[] for i in range(jma_anim.node_count)]

    # skip the first frame for overlays, otherwise the last frame
    last_keyframe = jma_anim.frame_count - 2
    skip_last_frame = (jma_anim.anim_type != "overlay")

    # calculate the keyframes and their counts based on the keyframes
    # in the jma_anim as well as the node transform flags
    for ni in range(jma_anim.node_count):
        # make sure to remove keyframe 0 since its stored in the default data
        if rot_flags[ni]:
            rot_keyframes_by_nodes[ni] = make_clean_keyframes_copy(
                jma_anim.rot_keyframes[ni],
                skip_last_frame, last_keyframe)

        if trans_flags[ni]:
            trans_keyframes_by_nodes[ni] = make_clean_keyframes_copy(
                jma_anim.trans_keyframes[ni],
                skip_last_frame, last_keyframe)

        if scale_flags[ni]:
            scale_keyframes_by_nodes[ni] = make_clean_keyframes_copy(
                jma_anim.scale_keyframes[ni],
                skip_last_frame, last_keyframe)

    # make the keyframe arrays big enough to fill in the keyframe numbers
    rot_keyframes.append(0)
    trans_keyframes.append(0)
    scale_keyframes.append(0)
    rot_keyframes   *= sum(len(kfs) for kfs in rot_keyframes_by_nodes)
    trans_keyframes *= sum(len(kfs) for kfs in trans_keyframes_by_nodes)
    scale_keyframes *= sum(len(kfs) for kfs in scale_keyframes_by_nodes)

    # make the default data arrays big enough to fill in the default data
    rot_def_data.append(0)
    trans_def_data.append(0)
    scale_def_data.append(0)
    rot_def_data   *= 3 * jma_anim.node_count
    trans_def_data *= 3 * jma_anim.node_count
    scale_def_data *= sum(bool(f) for f in scale_flags)

    # make the frame data arrays big enough to fill in the frame data
    rot_keyframe_data.append(0)
    trans_keyframe_data.append(0)
    scale_keyframe_data.append(0)
    rot_keyframe_data   *= 3 * len(rot_keyframes)
    trans_keyframe_data *= 3 * len(trans_keyframes)
    scale_keyframe_data *= len(scale_keyframes)


    comp_quat = compression.compress_quaternion48
    sqrt = math.sqrt

    ri = ti = si = 0
    def_ri = def_ti = def_si = 0

    def_frame = jma_anim.frames[0]

    for ni in range(jma_anim.node_count):
        curr_rot_kfs   = array(rot_keyframes.typecode,
                               rot_keyframes_by_nodes[ni])
        curr_trans_kfs = array(trans_keyframes.typecode,
                               trans_keyframes_by_nodes[ni])
        curr_scale_kfs = array(scale_keyframes.typecode,
                               scale_keyframes_by_nodes[ni])

        keyframes_sane = True
        if curr_rot_kfs:
            keyframes_sane &= last_keyframe in curr_rot_kfs

        if curr_trans_kfs:
            keyframes_sane &= last_keyframe in curr_trans_kfs

        if curr_scale_kfs:
            keyframes_sane &= last_keyframe in curr_scale_kfs

        assert keyframes_sane, (
            "Compressed animations must contain either no "
            "keyframes, or at least the last stored keyframe."
            )

        # copy the keyframe indices for this node into the block
        rot_keyframes[ri: ri + len(curr_rot_kfs)] = curr_rot_kfs
        trans_keyframes[ti: ti + len(curr_trans_kfs)] = curr_trans_kfs
        scale_keyframes[si: si + len(curr_scale_kfs)] = curr_scale_kfs

        # fill in the default data for this node
        def_node_state = def_frame[ni]

        w0, w1, w2 = comp_quat(
            def_node_state.rot_i, def_node_state.rot_j,
            def_node_state.rot_k, def_node_state.rot_w)
        rot_def_data[def_ri] = w0
        rot_def_data[def_ri + 1] = w1
        rot_def_data[def_ri + 2] = w2
        def_ri += 3

        trans_def_data[def_ti] = def_node_state.pos_x / 100
        trans_def_data[def_ti + 1] = def_node_state.pos_y / 100
        trans_def_data[def_ti + 2] = def_node_state.pos_z / 100
        def_ti += 3

        if scale_flags[ni]:
            # only write a default scale if scale is animated
            scale_def_data[def_si] = def_node_state.scale
            def_si += 1

        # fill in the keyframe data for this node
        for kfi in curr_rot_kfs:
            node_state = jma_anim.frames[kfi][ni]
            w0, w1, w2 = comp_quat(
                node_state.rot_i, node_state.rot_j,
                node_state.rot_k, node_state.rot_w)
            rot_keyframe_data[3*ri] = w0
            rot_keyframe_data[3*ri + 1] = w1
            rot_keyframe_data[3*ri + 2] = w2

            ri += 1

        for kfi in curr_trans_kfs:
            node_state = jma_anim.frames[kfi][ni]
            trans_keyframe_data[3*ti] = node_state.pos_x / 100
            trans_keyframe_data[3*ti + 1] = node_state.pos_y / 100
            trans_keyframe_data[3*ti + 2] = node_state.pos_z / 100
            ti += 1

        for kfi in curr_scale_kfs:
            scale_keyframe_data[si] = jma_anim.frames[kfi][ni].scale
            si += 1


    # setup the keyframe counts, offsets, data stream offsets
    calc_keyframe_header_data(
        comp_anim_block,
        [len(rot_keyframes_by_nodes[i]) for i in
         range(len(rot_keyframes_by_nodes)) if rot_flags[i]],

        [len(trans_keyframes_by_nodes[i]) for i in
         range(len(trans_keyframes_by_nodes)) if trans_flags[i]],

        [len(scale_keyframes_by_nodes[i]) for i in
         range(len(scale_keyframes_by_nodes)) if scale_flags[i]],
        )

    #print("RECOMP")
    #print(comp_anim_block)
    return comp_anim_block.serialize(calc_pointers=False)


def deserialize_compressed_frame_data(anim):
    rot_keyframes_by_nodes = []
    trans_keyframes_by_nodes = []
    scale_keyframes_by_nodes = []

    keyframes = (rot_keyframes_by_nodes,
                 trans_keyframes_by_nodes,
                 scale_keyframes_by_nodes)
    if not anim.flags.compressed_data:
        return keyframes, ()

    # make a bunch of frames we can fill in below
    frames = [[JmaNodeState() for n in range(anim.node_count)]
              for f in range(anim.frame_count + 1)]

    rot_flags, trans_flags, scale_flags = get_anim_flags(anim)

    comp_anim_block = compressed_frames_def.build(
        rawdata=anim.frame_data.STEPTREE,
        root_offset=anim.offset_to_compressed_data)
    #print("ORIG")
    #print(comp_anim_block)

    # get the keyframe counts and keyframe offsets
    rot_keyframe_headers   = [(v & 4095, v >> 12) for v in
                              comp_anim_block.rotation.keyframe_head]
    trans_keyframe_headers = [(v & 4095, v >> 12) for v in
                              comp_anim_block.translation.keyframe_head]
    scale_keyframe_headers = [(v & 4095, v >> 12) for v in
                              comp_anim_block.scale.keyframe_head]

    rot_keyframes   = comp_anim_block.rotation.keyframes
    trans_keyframes = comp_anim_block.translation.keyframes
    scale_keyframes = comp_anim_block.scale.keyframes

    rot_def_data   = comp_anim_block.rotation.default_data
    trans_def_data = comp_anim_block.translation.default_data
    scale_def_data = comp_anim_block.scale.default_data

    rot_keyframe_data   = comp_anim_block.rotation.keyframe_data
    trans_keyframe_data = comp_anim_block.translation.keyframe_data
    scale_keyframe_data = comp_anim_block.scale.keyframe_data

    decomp_quat = compression.decompress_quaternion48
    blend_trans = matrices.lerp_blend_vectors
    blend_quats = matrices.nlerp_blend_quaternions
    sqrt = math.sqrt

    ri = ti = si = 0
    for ni in range(anim.node_count):
        rot_kf_ct  = trans_kf_ct  = scale_kf_ct  = 0
        rot_kf_off = trans_kf_off = scale_kf_off = 0

        rot_def = decomp_quat(*rot_def_data[3 * ni: 3 * (ni + 1)])
        trans_def = trans_def_data[3 * ni: 3 * (ni + 1)]
        scale_def = 1.0

        if rot_flags[ni]:
            rot_kf_ct, rot_kf_off = rot_keyframe_headers[ri]
            ri += 1

        if trans_flags[ni]:
            trans_kf_ct, trans_kf_off = trans_keyframe_headers[ti]
            ti += 1

        if scale_flags[ni]:
            scale_kf_ct, scale_kf_off = scale_keyframe_headers[si]
            si += 1
            scale_def = scale_def_data[ni]

        # add this nodes keyframes to the keyframe lists in the jma_anim
        for kf_ct, kf_off, all_kfs, kfs_by_nodes in (
                (rot_kf_ct, rot_kf_off, rot_keyframes,
                 rot_keyframes_by_nodes),
                (trans_kf_ct, trans_kf_off, trans_keyframes,
                 trans_keyframes_by_nodes),
                (scale_kf_ct, scale_kf_off, scale_keyframes,
                 scale_keyframes_by_nodes)):
            kfs_by_nodes.append(list(all_kfs[kf_off: kf_off + kf_ct]))


        if rot_kf_ct:
            rot_first_kf = rot_keyframes[rot_kf_off]
            rot_last_kf  = rot_keyframes[rot_kf_off + rot_kf_ct - 1]
            rot_first = decomp_quat(*rot_keyframe_data[
                3 * rot_kf_off:
                3 * (rot_kf_off + 1)])
            rot_last = decomp_quat(*rot_keyframe_data[
                3 * (rot_kf_off + rot_kf_ct - 1):
                3 * (rot_kf_off + rot_kf_ct)])

        if trans_kf_ct:
            trans_first_kf = trans_keyframes[trans_kf_off]
            trans_last_kf  = trans_keyframes[trans_kf_off + trans_kf_ct - 1]
            trans_first = trans_keyframe_data[
                3 * trans_kf_off:
                3 * (trans_kf_off + 1)]
            trans_last = trans_keyframe_data[
                3 * (trans_kf_off + trans_kf_ct - 1):
                3 * (trans_kf_off + trans_kf_ct)]

        if scale_kf_ct:
            scale_first_kf = scale_keyframes[scale_kf_off]
            scale_last_kf  = scale_keyframes[scale_kf_off + scale_kf_ct - 1]
            scale_first = scale_keyframe_data[scale_kf_off]
            scale_last  = scale_keyframe_data[scale_kf_off + scale_kf_ct - 1]

        for fi in range(anim.frame_count):
            node_frame = frames[fi][ni]

            if not rot_kf_ct or fi == 0:
                # first frame OR only default data stored for this node
                i, j, k, w = rot_def
            elif fi == rot_last_kf:
                # frame is the last keyframe. repeat it to the end
                i, j, k, w = rot_last
            elif fi < rot_first_kf:
                # frame is before the first stored keyframe.
                # blend from default data to first keyframe.
                i, j, k, w = blend_quats(
                    rot_def, rot_first, fi / rot_first_kf)
            else:
                # frame is at/past the first stored keyframe.
                # don't need to use default data at all.
                kf_i = get_keyframe_index_of_frame(
                    fi, rot_keyframes, rot_kf_ct, rot_kf_off)
                kf0 = rot_keyframes[kf_i]
                q0 = decomp_quat(
                    *rot_keyframe_data[kf_i * 3: (kf_i + 1) * 3])

                if fi == kf0:
                    # this keyframe is the frame we want.
                    # no blending required
                    i, j, k, w = q0
                else:
                    kf1 = rot_keyframes[kf_i + 1]
                    ratio = (fi - kf0) / (kf1 - kf0)
                    kf_i += 1
                    q1 = decomp_quat(
                        *rot_keyframe_data[kf_i * 3: (kf_i + 1) * 3])
                    i, j, k, w = blend_quats(q0, q1, ratio)


            if not trans_kf_ct or fi == 0:
                # first frame OR only default data stored for this node
                x, y, z = trans_def
            elif fi == trans_last_kf:
                # frame is the last keyframe. repeat it to the end
                x, y, z = trans_last
            elif fi < trans_first_kf:
                # frame is before the first stored keyframe.
                # blend from default data to first keyframe.
                x, y, z = blend_trans(
                    trans_def, trans_first, fi / trans_first_kf)
            else:
                # frame is at/past the first stored keyframe.
                # don't need to use default data at all.
                kf_i = get_keyframe_index_of_frame(
                    fi, trans_keyframes, trans_kf_ct, trans_kf_off)
                kf0 = trans_keyframes[kf_i]
                p0 = trans_keyframe_data[kf_i * 3: (kf_i + 1) * 3]

                if fi == kf0:
                    # this keyframe is the frame we want.
                    # no blending required
                    x, y, z = p0
                else:
                    kf1 = trans_keyframes[kf_i + 1]
                    ratio = (fi - kf0) / (kf1 - kf0)

                    kf_i += 1
                    p1 = trans_keyframe_data[kf_i * 3: (kf_i + 1) * 3]
                    x, y, z = blend_trans(p0, p1, ratio)


            if not scale_kf_ct or fi == 0:
                # first frame OR only default data stored for this node
                scale = scale_def
            elif fi == scale_last_kf:
                # frame is the last keyframe. repeat it to the end
                scale = scale_last
            elif fi < scale_first_kf:
                # frame is before the first stored keyframe.
                # blend from default data to first keyframe.
                ratio = fi / scale_first_kf
                scale = scale_def * (1 - ratio) + scale_first * ratio
            else:
                # frame is at/past the first stored keyframe.
                # don't need to use default data at all.
                kf_i = get_keyframe_index_of_frame(
                    fi, scale_keyframes, scale_kf_ct, scale_kf_off)

                if fi == kf0:
                    # this keyframe is the frame we want.
                    # no blending required
                    scale = scale_keyframes[kf_i]
                else:
                    ratio = ((fi - scale_keyframes[kf_i]) /
                             (scale_keyframes[kf_i + 1] -
                              scale_keyframes[kf_i]))
                    scale = (
                        scale_keyframe_data[kf_i] * (1 - ratio) +
                        scale_keyframe_data[kf_i + 1] * ratio)


            nmag = i**2 + j**2 + k**2 + w**2
            if nmag:
                nmag = 1 / sqrt(nmag)
                node_frame.rot_i = i * nmag
                node_frame.rot_j = j * nmag
                node_frame.rot_k = k * nmag
                node_frame.rot_w = w * nmag

            node_frame.pos_x = x * 100
            node_frame.pos_y = y * 100
            node_frame.pos_z = z * 100

            node_frame.scale = scale

    if anim.type.enum_name != "overlay":
        # duplicate the first frame to the last frame for non-overlays
        frames[-1] = deepcopy(frames[0])

    return keyframes, frames


def get_keyframe_index_of_frame(frame, keyframes,
                                keyframe_count=None, offset=0):
    if keyframe_count is None:
        keyframe_count = len(keyframes) - offset

    # TODO: make this more efficent using a binary search
    for i in range(offset, offset + keyframe_count - 1):
        if keyframes[i] <= frame and frame < keyframes[i + 1]:
            return i

    raise ValueError(
        "No keyframes pairs containing frame %s" % frame)


def get_anim_flags(anim):
    rot_flags   = anim.rot_flags0   | (anim.rot_flags1 << 32)
    trans_flags = anim.trans_flags0 | (anim.trans_flags1 << 32)
    scale_flags = anim.scale_flags0 | (anim.scale_flags1 << 32)

    rot_flags   = [bool(rot_flags   & (1 << i)) for i in range(anim.node_count)]
    trans_flags = [bool(trans_flags & (1 << i)) for i in range(anim.node_count)]
    scale_flags = [bool(scale_flags & (1 << i)) for i in range(anim.node_count)]
    return rot_flags, trans_flags, scale_flags


def make_clean_keyframes_copy(keyframes, skip_last_frame, last_keyframe):
    kfs = list(keyframes)
    if not kfs:
        return kfs

    if kfs[0] == 0:
        kfs.pop(0)

    if kfs and skip_last_frame and kfs[-1] == last_keyframe + 1:
        kfs.pop(-1)

    return kfs


def calc_keyframe_header_data(comp_anim_block, rot_keyframe_counts,
                              trans_keyframe_counts, scale_keyframe_counts):

    for keyframe_counts, keyframe_header in (
            (rot_keyframe_counts,   comp_anim_block.rotation.keyframe_head),
            (trans_keyframe_counts, comp_anim_block.translation.keyframe_head),
            (scale_keyframe_counts, comp_anim_block.scale.keyframe_head)):
        if not keyframe_counts:
            pass
        elif max(keyframe_counts) >= 4096:
            raise ValueError(
                "Too many keyframes to compress per node. Must be < 4096, "
                "but got %s" % max(keyframe_counts))
        elif sum(keyframe_counts) >= 1048576:
            raise ValueError(
                "Too many keyframes to compress in total. Must be < 1048576 "
                "but got %s" % sum(keyframe_counts))

        keyframe_header.append(0)
        keyframe_header *= len(keyframe_counts)

        off = i = 0
        for ct in keyframe_counts:
            keyframe_header[i] = ct | (off << 12)
            off += ct
            i += 1

    # setup the data stream offsets
    off = None
    for offs, data in (
            (comp_anim_block.rotation_offsets, comp_anim_block.rotation),
            (comp_anim_block.translation_offsets, comp_anim_block.translation),
            (comp_anim_block.scale_offsets, comp_anim_block.scale)):
        for i in range(4):
            if off is None:
                off = 44
            else:
                offs[i] = off

            off += len(data[i]) * data[i].itemsize



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag

class PphyTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)
        tagdata = self.data.tagdata

        d = tagdata.density
        tagdata.scaled_density = d * 355840 / 3

        try:
            tagdata.water_gravity_scale = -((d - 1) / (d + 1))
        except ZeroDivisionError:
            tagdata.water_gravity_scale = float("-inf")

        # 0.0011 is the density of air relative to water
        try:
            tagdata.air_gravity_scale = -((d / 0.0011 - 1) /
                                          (d / 0.0011 + 1))
        except ZeroDivisionError:
            tagdata.air_gravity_scale = float("-inf")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'JmsMaterial', )


class JmsMaterial:
    __slots__ = (
        "name", "tiff_path",
        "shader_path", "shader_type",
        "properties"
        )
    def __init__(self, name="__unnamed", tiff_path="<none>",
                 shader_path="", shader_type="", properties=""):
        for c in "!@#$%^&*-.":
            if c in name and c not in properties:
                properties += c
            name = name.replace(c, '')

        self.name = name
        self.tiff_path = tiff_path
        self.shader_path = shader_path if shader_path else name
        self.shader_type = shader_type
        self.properties = properties

    @property
    def ai_defeaning(self): return "&" in self.properties
    @ai_defeaning.setter
    def ai_defeaning(self, new_val):
        self.properties = self.properties.replace("&", "") + ("&" if new_val else "")

    @property
    def allow_transparency(self): return "#" in self.properties
    @allow_transparency.setter
    def allow_transparency(self, new_val):
        self.properties = self.properties.replace("#", "") + ("#" if new_val else "")

    @property
    def breakable(self): return "-" in self.properties
    @breakable.setter
    def breakable(self, new_val):
        self.properties = self.properties.replace("-", "") + ("-" if new_val else "")

    # collision_only means the player collides, but not projectiles
    @property
    def collision_only(self): return "@" in self.properties
    @collision_only.setter
    def collision_only(self, new_val):
        self.properties = self.properties.replace("@", "") + ("@" if new_val else "")

    @property
    def double_sided(self): return "%" in self.properties
    @double_sided.setter
    def double_sided(self, new_val):
        self.properties = self.properties.replace("%", "") + ("%" if new_val else "")

    @property
    def exact_portal(self): return "." in self.properties
    @exact_portal.setter
    def exact_portal(self, new_val):
        self.properties = self.properties.replace(".", "") + ("." if new_val else "")

    @property
    def fog_plane(self): return "$" in self.properties
    @fog_plane.setter
    def fog_plane(self, new_val):
        self.properties = self.properties.replace("$", "") + ("$" if new_val else "")

    @property
    def ladder(self): return "^" in self.properties
    @ladder.setter
    def ladder(self, new_val):
        self.properties = self.properties.replace("^", "") + ("^" if new_val else "")

    # this is what sky and invisible collision get set to
    @property
    def large_collideable(self): return "*" in self.properties
    @large_collideable.setter
    def large_collideable(self, new_val):
        self.properties = self.properties.replace("*", "") + ("*" if new_val else "")

    @property
    def render_only(self): return "!" in self.properties
    @render_only.setter
    def render_only(self, new_val):
        self.properties = self.properties.replace("!", "") + ("!" if new_val else "")

    def __repr__(self):
        return """JmsMaterial(name=%s,
    tiff_path=%s
)""" % (self.name, self.tiff_path)
